"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

Robot arm problem from:
Liu, H., Xu, S., & Wang, X. Sampling strategies and metamodeling techniques for engineering design:
comparison and application. In ASME Turbo Expo 2016: Turbomachinery Technical Conference and Exposition.
American Society of Mechanical Engineers. June, 2016.

An, J., and Owen, A. Quasi-Regression. Journal of complexity, 17(4), pp. 588-607, 2001.
"""

import numpy as np

from smt.problems.problem import Problem


class RobotArm(Problem):
    def _initialize(self):
        self.options.declare("name", "RobotArm", types=str)
        self.options.declare("ndim", 2, types=int)

    def _setup(self):
        assert self.options["ndim"] % 2 == 0, "ndim must be divisible by 2"

        # Length L
        self.xlimits[0::2, 0] = 0.0
        self.xlimits[0::2, 1] = 1.0

        # Angle theta
        self.xlimits[1::2, 0] = 0.0
        self.xlimits[1::2, 1] = 2 * np.pi

    def _evaluate(self, x, kx):
        """
        Arguments
        ---------
        x : ndarray[ne, nx]
            Evaluation points.
        kx : int or None
            Index of derivative (0-based) to return values with respect to.
            None means return function value rather than derivative.

        Returns
        -------
        ndarray[ne, 1]
            Functions values if kx=None or derivative values if kx is an int.
        """
        ne, nx = x.shape

        nseg = int(self.options["ndim"] / 2)

        pos_x = np.zeros(ne, complex)
        pos_y = np.zeros(ne, complex)
        for iseg in range(nseg):
            L = x[:, 2 * iseg + 0]
            pos_x += L * np.cos(np.sum(x[:, 1 : 2 * iseg + 2 : 2], axis=1))
            pos_y += L * np.sin(np.sum(x[:, 1 : 2 * iseg + 2 : 2], axis=1))

        y = np.zeros((ne, 1), complex)
        d_pos_x = np.zeros(ne, complex)
        d_pos_y = np.zeros(ne, complex)

        # To remove warning: NaN can be ignored
        np.seterr(invalid="ignore")
        if kx is None:
            y[:, 0] = (pos_x**2 + pos_y**2) ** 0.5
        else:
            kseg = int(np.floor(kx / 2))
            if kx % 2 == 0:
                d_pos_x[:] += np.cos(np.sum(x[:, 1 : 2 * kseg + 2 : 2], axis=1))
                d_pos_y[:] += np.sin(np.sum(x[:, 1 : 2 * kseg + 2 : 2], axis=1))
                y[:, 0] += pos_x / (pos_x**2 + pos_y**2) ** 0.5 * d_pos_x
                y[:, 0] += pos_y / (pos_x**2 + pos_y**2) ** 0.5 * d_pos_y
            elif kx % 2 == 1:
                for iseg in range(nseg):
                    L = x[:, 2 * iseg + 0]
                    if kseg <= iseg:
                        d_pos_x[:] -= L * np.sin(
                            np.sum(x[:, 1 : 2 * iseg + 2 : 2], axis=1)
                        )
                        d_pos_y[:] += L * np.cos(
                            np.sum(x[:, 1 : 2 * iseg + 2 : 2], axis=1)
                        )
                y[:, 0] += pos_x / (pos_x**2 + pos_y**2) ** 0.5 * d_pos_x
                y[:, 0] += pos_y / (pos_x**2 + pos_y**2) ** 0.5 * d_pos_y

        return y



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

Base class for benchmarking/test problems.
"""

from typing import Optional

import numpy as np

from smt.utils.checks import ensure_2d_array
from smt.utils.design_space import BaseDesignSpace, DesignSpace
from smt.utils.options_dictionary import OptionsDictionary


class Problem:
    def __init__(self, **kwargs):
        """
        Constructor where values of options can be passed in.

        For the list of options, see the documentation for the problem being used.

        Parameters
        ----------
        **kwargs : named arguments
            Set of options that can be optionally set; each option must have been declared.

        Examples
        --------
        >>> from smt.problems import Sphere
        >>> prob = Sphere(ndim=3)
        """
        self.options = OptionsDictionary()
        self.options.declare("ndim", 1, types=int)
        self.options.declare("return_complex", False, types=bool)
        self._initialize()
        self.options.update(kwargs)

        self.xlimits = np.zeros((self.options["ndim"], 2))
        self._design_space = None

        self.eval_x = None
        self.eval_is_acting = None

        self._setup()

    def _initialize(self) -> None:
        """
        Implemented by problem to declare options (optional).

        Examples
        --------
        self.options.declare('option_name', default_value, types=(bool, int), desc='description')
        """
        pass

    def _setup(self) -> None:
        pass

    def _set_design_space(self, design_space: BaseDesignSpace):
        """
        Set the design space definition (best is to use the smt.utils.design_space.DesignSpace class directly) of
        this problem from the _setup function. If used, there is no need to set xlimits.
        """
        self._design_space = design_space
        self.options["ndim"] = len(design_space.design_variables)
        self.xlimits = design_space.get_num_bounds()

    @property
    def design_space(self) -> BaseDesignSpace:
        """Gets the design space definitions as an instance of BaseDesignSpace"""
        if self._design_space is None:
            self._design_space = DesignSpace(self.xlimits)
        return self._design_space

    def sample(self, n):
        x, _ = self.design_space.sample_valid_x(n)
        return x

    def __call__(self, x: np.ndarray, kx: Optional[int] = None) -> np.ndarray:
        """
        Evaluate the function.
        The input vectors might be corrected if it is a hierarchical design space. You can get the corrected x and
        information about which variables are acting from: problem.eval_x and problem.eval_is_acting

        Parameters
        ----------
        x : ndarray[n, nx] or ndarray[n]
            Evaluation points where n is the number of evaluation points.
        kx : int or None
            Index of derivative (0-based) to return values with respect to.
            None means return function value rather than derivative.

        Returns
        -------
        ndarray[n, 1]
            Functions values if kx=None or derivative values if kx is an int.
        """
        x = ensure_2d_array(x, "x")

        if x.shape[1] != self.options["ndim"]:
            raise ValueError(
                "The second dimension of x should be %i" % self.options["ndim"]
            )

        if kx is not None:
            if not isinstance(kx, int) or kx < 0:
                raise TypeError("kx should be None or a non-negative int.")

        # Correct the design vector and get information about which design variables are active
        x_corr, self.eval_is_acting = self.design_space.correct_get_acting(x)
        self.eval_x = x_corr

        y = self._evaluate(x_corr, kx)

        if self.options["return_complex"]:
            return y
        else:
            return np.real(y)

    def _evaluate(self, x: np.ndarray, kx: Optional[int] = None) -> np.ndarray:
        """
        Implemented by surrogate models to evaluate the function.

        Parameters
        ----------
        x : ndarray[n, nx]
            Evaluation points where n is the number of evaluation points.
        kx : int or None
            Index of derivative (0-based) to return values with respect to.
            None means return function value rather than derivative.

        Returns
        -------
        ndarray[n, 1]
            Functions values if kx=None or derivative values if kx is an int.
        """
        raise Exception("This problem has not been implemented correctly")



"""
Author: Dr. Mohamed A. Bouhlel <mbouhlel@umich.edu>

This package is distributed under New BSD license.
"""

import numpy as np

from sklearn.cross_decomposition import PLSRegression as pls

from smt.surrogate_models.krg_based import KrgBased
from smt.utils.kriging import componentwise_distance_PLS


class KPLS(KrgBased):
    name = "KPLS"

    def _initialize(self):
        super(KPLS, self)._initialize()
        declare = self.options.declare
        declare("n_comp", 1, types=int, desc="Number of principal components")
        # KPLS used only with "abs_exp" and "squar_exp" correlations
        declare(
            "corr",
            "squar_exp",
            values=("abs_exp", "squar_exp", "pow_exp"),
            desc="Correlation function type",
            types=(str),
        )
        declare(
            "eval_n_comp",
            False,
            types=(bool),
            values=(True, False),
            desc="n_comp evaluation flag",
        )
        declare(
            "eval_comp_treshold",
            1.0,
            types=(float),
            desc="n_comp evaluation treshold for Wold's R criterion",
        )
        declare(
            "cat_kernel_comps",
            None,
            types=list,
            desc="Number of components for PLS categorical kernel",
        )

    def _compute_pls(self, X, y):
        _pls = pls(self.options["n_comp"])
        self.coeff_pls = 0
        if np.shape(X)[0] < self.options["n_comp"] + 1:
            raise ValueError(
                "ValueError: The database should be at least "
                + str(self.options["n_comp"] + 1)
                + " points (currently "
                + str(np.shape(X)[0])
                + ")."
            )
        else:
            if np.shape(X)[1] == 1:
                self.coeff_pls = np.atleast_2d(np.array([1]))
            else:
                self.coeff_pls = abs(_pls.fit(X.copy(), y.copy()).x_rotations_)
        return X, y

    def _componentwise_distance(self, dx, opt=0, theta=None, return_derivative=False):
        d = componentwise_distance_PLS(
            dx,
            self.options["corr"],
            self.options["n_comp"],
            self.coeff_pls,
            power=self.options["pow_exp_power"],
            theta=theta,
            return_derivative=return_derivative,
        )
        return d

    def _estimate_number_of_components(self):
        """
        self.options[n_comp] value from user is ignored and replaced by an estimated one wrt Wold's R criterion.
        """
        eval_comp_treshold = self.options["eval_comp_treshold"]
        X = self.training_points[None][0][0]
        y = self.training_points[None][0][1]
        k_fold = 4
        nbk = int(self.nt / k_fold)
        press_m = 0.0
        press_m1 = 0.0
        self.options["n_comp"] = 0
        nextcomp = True
        while nextcomp:
            self.options["n_comp"] += 1
            press_m = press_m1
            press_m1 = 0
            self.options["theta0"] = [0.1]
            for fold in range(k_fold):
                self.nt = len(X) - nbk
                todel = np.arange(fold * nbk, (fold + 1) * nbk)
                Xfold = np.copy(X)
                Xfold = np.delete(X, todel, axis=0)
                yfold = np.copy(y)
                yfold = np.delete(y, todel, axis=0)
                Xtest = np.copy(X)[fold * nbk : (fold + 1) * nbk, :]
                ytest = np.copy(y)[fold * nbk : (fold + 1) * nbk, :]

                self.training_points[None][0][0] = Xfold
                self.training_points[None][0][1] = yfold
                try:
                    self._new_train()
                except ValueError:
                    self.options["n_comp"] -= 1
                    nextcomp = False
                    break
                ye = self._predict_values(Xtest)
                press_m1 = press_m1 + np.sum(np.power((1 / len(X)) * (ye - ytest), 2))
            if self.options["n_comp"] > 1 and press_m1 / press_m > eval_comp_treshold:
                self.options["n_comp"] -= 1
                nextcomp = False
        self.training_points[None][0][0] = X
        self.training_points[None][0][1] = y
        self.nt = len(X)
        self.options["theta0"] = [0.1]

    def _train(self):
        """
        Train the model
        """
        # outputs['sol'] = self.sol

        if self.options["eval_n_comp"]:
            self._estimate_number_of_components()
        self._new_train()



"""
Authors : Morgane Menz / Alexandre Thouvenot
Some parts are copied from KrgBased SMT class
"""

import numpy as np
from scipy import linalg

from smt.utils.kriging import differences


def covariance_matrix(krg, X, conditioned=True):
    """
    This function computes the covariance matrix (with conditioned kernel or not) at point(s) X.

    Parameters
    ----------
    krg : SMT KRG
        SMT kriging model already trained
    X : array_like
        Array with shape (n_eval, n_feature) composed of the point(s) on which the
        covariance matrix will be computed
    conditioned : Boolean
        Option to whether computes covariance matrix with conditioned kernel or not.

    Returns
    -------
    cov_matrix : array_like
        if conditioned: Array with shape(n_eval, n_eval) composed of the value(s) of the
        conditioned kernel
    cov_matrix : array_like
        if not conditioned: Array with shape(n_eval, n_eval) composed of the value(s) of the
        non conditioned kernel
    """
    X_cont = (X - krg.X_offset) / krg.X_scale
    d = differences(X_cont, Y=krg.X_norma.copy())
    cross_d = differences(X_cont, Y=X_cont)

    C = krg.optimal_par["C"]
    theta = krg.optimal_theta
    n_eval = X.shape[0]

    k = krg._correlation_types[krg.options["corr"]](
        theta, krg._componentwise_distance(cross_d)
    ).reshape(n_eval, n_eval)
    if not conditioned:
        cov_matrix = krg.optimal_par["sigma2"] * k
        return cov_matrix

    r = krg._correlation_types[krg.options["corr"]](
        theta, krg._componentwise_distance(d)
    ).reshape(n_eval, -1)
    rt = linalg.solve_triangular(C, r.T, lower=True)

    u = linalg.solve_triangular(
        krg.optimal_par["G"].T,
        np.dot(krg.optimal_par["Ft"].T, rt)
        - krg._regression_types[krg.options["poly"]](X_cont).T,
    )

    cov_matrix = krg.optimal_par["sigma2"] * (k - rt.T.dot(rt) + u.T.dot(u))

    return cov_matrix


def sample_trajectory(krg, X, n_traj, method="eigen", eps=10 ** (-10)):
    """
    This function samples gaussian process trajectories with eigen decomposition or Cholesky decomposition.

    Parameters
    ----------
    krg : SMT KRG
        SMT kriging model already trained
    X : array_like
        Array with shape (n_eval, n_feature) composed of the point(s) on which the
        trajctories will be sampled
    n_traj : int
        Number of sampled trajectories
    method : string
        Option to whether samples trajectories with eigen or Cholesky method. Cholesky decomposition
        might not be possible because of bad conditioning
    eps : float
        Threshold used to floor negative eigen value

    Returns
    -------
    traj : array_like
        Array with shape(n_eval, n_traj) composed of the sampled trajectorie with a chosen
        decomposition method
    """
    n_eval = X.shape[0]
    cov = covariance_matrix(krg, X)
    if method == "eigen":
        v, w = np.linalg.eigh(cov)
        v[np.abs(v) < eps] = np.zeros_like(v[np.abs(v) < eps])
        C = w.dot(np.diagflat(np.sqrt(v)))
    if method == "cholesky":
        C = np.linalg.cholesky(cov)
    mean_ = krg._predict_values(X).reshape(-1, 1)
    traj = (mean_ + C.dot(np.random.randn(n_eval, n_traj))).reshape(n_eval, n_traj)

    return traj


def gauss_legendre_grid(bounds, n_point):
    """
    This function creates a grid to computes integrals with Gauss-Legendre quadrature method.

    Parameters
    ----------
    bounds : array_like
        Array with shape (2, int_dims) where dims is the number of integration dimension.
        It iscontaining the integration domain where bounds[0] is the inferior integration
        bound and bounds[1] is the superior integration bound.
    n_point : int
        Number of point in each dimension

    Returns
    -------
    x_grid : array_like
        Array with shape(n_point**int_dims, int_dims) composed of point(s) on which the
        integral will be computed
    weights_grid : array_like
        Array with shape(n_point**int_dims, 1) composed of weigths used to computed the
        integral
    """
    dims = bounds.shape[1]
    x, weights = np.polynomial.legendre.leggauss(n_point)
    x_list = [
        (bounds[1][i] - bounds[0][i]) / 2 * x + (bounds[1][i] + bounds[0][i]) / 2
        for i in range(dims)
    ]
    x_grid = np.meshgrid(*x_list)
    x_grid = np.concatenate([x_grid[i].reshape(-1, 1) for i in range(dims)], axis=1)
    weights_grid = np.meshgrid(*[weights] * dims)
    weights_grid = np.prod(weights_grid, axis=0).reshape(-1, 1)
    return x_grid, weights_grid


def rectangular_grid(bounds, n_point):
    """
    This function creates a grid to computes integrals with rectangular grid method.

    Parameters
    ----------
    bounds : array_like
        Array with shape (2, int_dims) where dims is the number of integration dimension.
        It iscontaining the integration domain where bounds[0] is the inferior integration
        bound and bounds[1] is the superior integration bound.
    n_point : int
        Number of point in each dimension

    Returns
    -------
    x_grid : array_like
        Array with shape(n_point**int_dims, int_dims) composed of point(s) on which the
        integral will be computed
    weights_grid : array_like
        Array with shape(n_point**int_dims, 1) composed of weigths used to computed the
        integral
    """
    dims = bounds.shape[1]
    x, weights = np.linspace(-1, 1, n_point), np.ones(n_point) / n_point
    x_list = [
        (bounds[1][i] - bounds[0][i]) / 2 * x + (bounds[1][i] + bounds[0][i]) / 2
        for i in range(dims)
    ]
    x_grid = np.meshgrid(*x_list)
    x_grid = np.concatenate([x_grid[i].reshape(-1, 1) for i in range(dims)], axis=1)
    weights_grid = np.meshgrid(*[weights] * dims)
    weights_grid = np.prod(weights_grid, axis=0).reshape(-1, 1)
    return x_grid, weights_grid


def simpson_weigths(n_points, h):
    """
    This function computes Simpson quadrature weigths in one dimension.

    Parameters
    ----------
    n_point : int
        Number of weigths
    h : float
        Scaling coefficient

    Returns
    -------
    weights : array_like
        Array with shape(n_point,) composed of weigths used to computed the
        integral
    """
    weights = np.zeros(n_points)
    for i in range((n_points + 1) // 2):
        weights[2 * i] = 2.0 * h / 3.0
        weights[2 * i - 1] = 4.0 * h / 3.0
    weights[0], weights[-1] = h / 3.0, h / 3.0
    return weights


def simpson_grid(bounds, n_point):
    """
    This function creates a grid to computes integrals with Simpson quadrature.

    Parameters
    ----------
    bounds : array_like
        Array with shape (2, int_dims) where dims is the number of integration dimension.
        It iscontaining the integration domain where bounds[0] is the inferior integration
        bound and bounds[1] is the superior integration bound.
    n_point : int
        Number of point in each dimension

    Returns
    -------
    x_grid : array_like
        Array with shape (n_point**int_dims, int_dims) composed of point(s) on which the
        integral will be computed
    weights_grid : array_like
        Array with shape (n_point**int_dims, 1) composed of weigths used to computed the
        integral
    """
    dims = bounds.shape[1]
    x = np.linspace(-1, 1, n_point)
    x_list = [
        (bounds[1][i] - bounds[0][i]) / 2 * x + (bounds[1][i] + bounds[0][i]) / 2
        for i in range(dims)
    ]
    x_grid = np.meshgrid(*x_list)
    x_grid = np.concatenate([x_grid[i].reshape(-1, 1) for i in range(dims)], axis=1)
    weights_grid = np.meshgrid(
        *[
            simpson_weigths(n_point, (bounds[1][i] - bounds[0][i]) / n_point)
            for i in range(dims)
        ]
    )
    weights_grid = np.prod(weights_grid, axis=0).reshape(-1, 1)
    return x_grid, weights_grid


def eig_grid(krg, x_grid, weights_grid):
    """
    This function computes eigen values and eigen vectors of Karhunen-Loève decomposition
    with Nyström method on a given interpolation.

    Parameters
    ----------
    krg : SMT KRG
        SMT kriging model already trained
    x_grid : array_like
        Array with shape (n_point, n_features) composed of point(s) on which the
        integral will be computed
    weights_grid : array_like
        Array with shape (n_point, 1) composed of weigths used to computed the
        integral

    Returns
    -------
    eig_val : array_like
        Eigen values
    eig_vec : array_like
        Eigen Vectors
    M : int
        Number of retained eigen values

    """
    C = covariance_matrix(krg, x_grid, conditioned=False)

    W_sqrt = np.sqrt(np.diagflat(weights_grid))
    B = W_sqrt.dot(C).dot(W_sqrt)
    eig_val, eig_vec = np.linalg.eigh(B)

    ind = (-eig_val).argsort()
    eig_val = eig_val[ind]
    eig_vec = eig_vec[:, ind]

    crit = 1.0 - np.cumsum(eig_val) / eig_val.sum()
    M = int(np.argwhere(crit > 10 ** (-8))[-1].item()) + 1

    return eig_val, eig_vec, M


def evaluate_eigen_function(krg, X, eig_val, eig_vec, x_grid, weights_grid, M):
    """
    This function evaluates eigen functions of  Karhunen-Loève decomposition
    with Nyström method on X.

    Parameters
    ----------
    krg : SMT KRG
        SMT kriging model already trained
    X : array_like
        Array with shape (n_eval, n_feature) containing point(s) on which the eigen
        function will be evaluated
    eig_val : array_like
        Eigen values
    eig_vec : array_like
        Eigen Vectors
    x_grid : array_like
        Array with shape (n_point, n_features) composed of point(s) on which the
        integral will be computed
    weights_grid : array_like
        Array with shape (n_point, 1) composed of weigths used to computed the
        integral
    M : int
        Number of retained eigen values

    Returns
    -------
    phi : array_like
        Value of retained eigen functions on X

    """
    W_sqrt = np.sqrt(np.diagflat(weights_grid))
    X_ = np.concatenate((krg.X_train, X), axis=0)
    n_X = X_.shape[0]
    n_grid = x_grid.shape[0]

    X_cont = (X_ - krg.X_offset) / krg.X_scale
    X_grid_cont = (x_grid - krg.X_offset) / krg.X_scale
    cross_d = (
        np.tile(X_cont, (n_grid, 1)) - X_grid_cont.repeat(repeats=n_X, axis=0)
    ) ** 2

    C = krg.optimal_par["sigma2"] * krg._correlation_types[krg.options["corr"]](
        krg.optimal_theta, cross_d
    ).reshape(n_grid, n_X)

    U = np.diagflat(np.sqrt(1 / eig_val[:M]))
    phi = U.dot(eig_vec[:, :M].T.dot(W_sqrt).dot(C))
    return phi


def sample_eigen(krg, X, eig_val, eig_vec, x_grid, weights_grid, M, n_traj):
    """
    This function samples trajectories of gaussian process with Karhunen-Loève decomposition
    with Nyström method.

    Parameters
    ----------
    krg : SMT KRG
        SMT kriging model already trained
    X : array_like
        Array with shape (n_eval, n_feature) containing point(s) on which the eigen
        functions will be evaluated and on which the trajectories will be sampled
    eig_val : array_like
        Eigen values
    eig_vec : array_like
        Eigen Vectors
    x_grid : array_like
        Array with shape (n_point, n_features) composed of point(s) on which the
        integral will be computed
    weights_grid : array_like
        Array with shape (n_point, 1) composed of weigths used to computed the
        integral
    M : int
        Number of retained eigen values
    n_traj : int
        Number trajectories

    Returns
    -------
    traj : array_like
        Trajectories sampled on X

    """
    phi = evaluate_eigen_function(krg, X, eig_val, eig_vec, x_grid, weights_grid, M)
    sample = np.random.randn(n_traj, M).dot(phi)
    Y = sample[:, : krg.X_train.shape[0]]
    Y_mean, Y_std = Y.mean(axis=1).reshape(-1, 1), Y.std(axis=1).reshape(-1, 1)
    Y_norma = ((Y - Y_mean) / Y_std).T

    C, Q, G, Ft = (
        krg.optimal_par["C"],
        krg.optimal_par["Q"],
        krg.optimal_par["G"],
        krg.optimal_par["Ft"],
    )
    Yt = linalg.solve_triangular(C, Y_norma, lower=True)
    beta = linalg.solve_triangular(G, np.dot(Q.T, Yt))

    rho = Yt - np.dot(Ft, beta)
    gamma = linalg.solve_triangular(C.T, rho)

    X_cont = (X - krg.X_offset) / krg.X_scale
    dx = differences(X_cont, Y=krg.X_norma.copy())
    d = krg._componentwise_distance(dx)
    r = krg._correlation_types[krg.options["corr"]](krg.optimal_theta, d).reshape(
        X.shape[0], krg.nt
    )
    y = np.zeros(X.shape[0])
    f = krg._regression_types[krg.options["poly"]](X_cont)
    y_ = np.dot(f, beta) + np.dot(r, gamma)
    y = Y_mean.flatten() + y_ * Y_std.flatten()

    mean = krg._predict_values(X).reshape(-1, 1) - y
    traj = mean + sample[:, krg.X_train.shape[0] :].T

    return traj



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.
"""

import unittest

import numpy as np


class SMTestCase(unittest.TestCase):
    def assert_error(self, computed, desired, atol=1e-15, rtol=1e-15):
        """
        Check relative error of a scalar or array.

        Parameters
        ----------
        computed : float or ndarray
            Computed value; should be the same type and shape as desired.
        desired : float or ndarray
            Desired value; should be the same type and shape as computed.
        atol : float
            Acceptable absolute error. Default is 1e-15.
        rtol : float
            Acceptable relative error. Default is 1e-15.
        """
        abs_error = np.linalg.norm(computed - desired)
        if np.linalg.norm(desired) > 0:
            rel_error = abs_error / np.linalg.norm(desired)
        else:
            rel_error = abs_error
        if abs_error > atol and rel_error > rtol:
            self.fail(
                "computed %s, desired %s, abs error %s, rel error %s, atol %s, rtol %s"
                % (
                    np.linalg.norm(computed),
                    np.linalg.norm(desired),
                    abs_error,
                    rel_error,
                    atol,
                    rtol,
                )
            )



"""
Authors: Nathalie Bartoli, Remy Priem, Remi Lafage, Emile Roux <emile.roux@univ-smb.fr>

This package is distributed under New BSD license.

"""

from types import FunctionType

import numpy as np
from scipy.optimize import minimize
from scipy.stats import norm

from smt.applications.application import SurrogateBasedApplication
from smt.applications.mixed_integer import (
    MixedIntegerContext,
    MixedIntegerSamplingMethod,
)
from smt.sampling_methods import LHS
from smt.surrogate_models import GEKPLS, KPLS, KPLSK, KRG, MGP
from smt.utils.design_space import (
    BaseDesignSpace,
    DesignSpace,
)


class Evaluator(object):
    """
    An interface for evaluation of a function at x points (nsamples of dimension nx).
    User can derive this interface and override the run() method to implement custom multiprocessing.
    """

    def run(self, fun, x):
        """
        Evaluates fun at x.

        Parameters
        ---------
        fun : function to evaluate: (nsamples, nx) -> (nsample, 1)

        x : np.ndarray[nsamples, nx]
            nsamples points of nx dimensions.

        Returns
        -------
        np.ndarray[nsample, 1]
            fun evaluations at the nsamples points.

        """
        return fun(x)


class EGO(SurrogateBasedApplication):
    def _initialize(self):
        super(EGO, self)._initialize()
        declare = self.options.declare

        declare("fun", None, types=FunctionType, desc="Function to minimize")
        declare(
            "criterion",
            "EI",
            types=str,
            values=["EI", "SBO", "LCB"],
            desc="criterion for next evaluation point determination: Expected Improvement, \
            Surrogate-Based Optimization or Lower Confidence Bound",
        )
        declare("n_iter", None, types=int, desc="Number of optimizer steps")
        declare(
            "n_max_optim",
            20,
            types=int,
            desc="Maximum number of internal optimizations",
        )
        declare("n_start", 20, types=int, desc="Number of optimization start points")
        declare(
            "n_parallel",
            1,
            types=int,
            desc="Number of parallel samples to compute using qEI criterion",
        )
        declare(
            "qEI",
            "KBLB",
            types=str,
            values=["KB", "KBLB", "KBUB", "KBRand", "CLmin"],
            desc="Approximated q-EI maximization strategy",
        )
        declare(
            "evaluator",
            default=Evaluator(),
            types=Evaluator,
            desc="Object used to run function fun to optimize at x points (nsamples, nxdim)",
        )
        declare(
            "n_doe",
            None,
            types=int,
            desc="Number of points of the initial LHS doe, only used if xdoe is not given",
        )
        declare("xdoe", None, types=np.ndarray, desc="Initial doe inputs")
        declare("ydoe", None, types=np.ndarray, desc="Initial doe outputs")
        declare("verbose", False, types=bool, desc="Print computation information")
        declare(
            "enable_tunneling",
            False,
            types=bool,
            desc="Enable the penalization of points that have been already evaluated in EI criterion",
        )
        declare(
            "surrogate",
            KRG(print_global=False),
            types=(KRG, KPLS, KPLSK, GEKPLS, MGP),
            desc="SMT kriging-based surrogate model used internaly",
        )
        self.options.declare(
            "random_state",
            types=(type(None), int, np.random.RandomState),
            desc="Numpy RandomState object or seed number which controls random draws",
        )

    def optimize(self, fun):
        """
        Optimizes fun

        Parameters
        ----------

        fun: function to optimize: ndarray[n, nx] or ndarray[n] -> ndarray[n, 1]

        Returns
        -------

        [nx, 1]: x optimum
        [1, 1]: y optimum
        int: index of optimum in data arrays
        [ndoe + n_iter, nx]: coord-x data
        [ndoe + n_iter, 1]: coord-y data
        """
        x_data, y_data = self._setup_optimizer(fun)
        n_iter = self.options["n_iter"]
        n_parallel = self.options["n_parallel"]

        for k in range(n_iter):
            # Virtual enrichement loop
            for p in range(n_parallel):
                # find next best x-coord point to evaluate
                x_et_k, success = self._find_best_point(
                    x_data, y_data, self.options["enable_tunneling"]
                )
                if not success:
                    self.log(
                        "Internal optimization failed at EGO iter = {}.{}".format(k, p)
                    )
                    break
                elif success:
                    self.log(
                        "Internal optimization succeeded at EGO iter = {}.{}".format(
                            k, p
                        )
                    )
                # Set temporaly the y-coord point based on the kriging prediction
                x_et_k = np.atleast_2d(x_et_k)
                if self.mixint:
                    x_et_k, _ = self.design_space.correct_get_acting(x_et_k)
                y_et_k = self._get_virtual_point(x_et_k, y_data)

                # Update y_data with predicted value
                y_data = y_data.reshape(y_data.shape[0], self.gpr.ny)
                y_data = np.vstack((y_data, y_et_k))
                x_data = np.atleast_2d(np.append(x_data, x_et_k, axis=0))

            # Compute the real values of y_data
            x_to_compute = np.atleast_2d(x_data[-n_parallel:])
            y = self._evaluator.run(fun, x_to_compute)
            y_data[-n_parallel:] = y

        # Find the optimal point
        ind_best = np.argmin(y_data if y_data.ndim == 1 else y_data[:, 0])
        x_opt = x_data[ind_best]
        y_opt = y_data[ind_best]
        return x_opt, y_opt, ind_best, x_data, y_data

    def log(self, msg):
        if self.options["verbose"]:
            print(msg)

    def EI(self, points, enable_tunneling=False, x_data=None):
        """Expected improvement"""
        y_data = np.atleast_2d(self.gpr.training_points[None][0][1])
        f_min = y_data[np.argmin(y_data[:, 0])]
        pred = self.gpr.predict_values(points)
        sig = np.sqrt(self.gpr.predict_variances(points))
        args0 = (f_min - pred) / sig
        args1 = (f_min - pred) * norm.cdf(args0)
        args2 = sig * norm.pdf(args0)
        if sig.size == 1 and sig == 0.0:  # can be use only if one point is computed
            return 0.0
        ei = args1 + args2
        # penalize the points already evaluated with tunneling
        if enable_tunneling:
            for i in range(len(points)):
                p = np.atleast_2d(points[i])
                EIp = self.EI(p, enable_tunneling=False)
                for x in x_data:
                    x = np.atleast_2d(x)
                    # if np.abs(p-x)<1:
                    # ei[i]=ei[i]*np.reciprocal(1+100*np.exp(-np.reciprocal(1-np.square(p-x))))
                    pena = (EIp - self.EI(x, enable_tunneling=False)) / (
                        1e-9 + np.power(np.linalg.norm(p - x), 4)
                    )
                    if pena > 0:
                        ei[i] = ei[i] - pena
                    ei[i] = max(ei[i], 0)
        return ei

    def SBO(self, point):
        """Surrogate based optimization: min the surrogate model by suing the mean mu"""
        res = self.gpr.predict_values(point)
        return res

    def LCB(self, point):
        """Lower confidence bound optimization: minimize by using mu - 3*sigma"""
        pred = self.gpr.predict_values(point)
        var = self.gpr.predict_variances(point)
        res = pred - 3.0 * np.sqrt(var)
        return res

    def _setup_optimizer(self, fun):
        """
        Instanciate internal surrogate used for optimization
        and setup function evaluator wrt options

        Parameters
        ----------

        fun: function to optimize: ndarray[n, nx] or ndarray[n] -> ndarray[n, 1]

        Returns
        -------

        ndarray: initial coord-x doe
        ndarray: initial coord-y doe = fun(xdoe)

        """
        # Set the model
        self.gpr = self.options["surrogate"]
        self.design_space: BaseDesignSpace = self.gpr.design_space
        if isinstance(self.design_space, DesignSpace):
            self.design_space.seed = self.options["random_state"]

        # Handle mixed integer optimization
        is_continuous = self.design_space.is_all_cont
        if not is_continuous:
            self.categorical_kernel = self.gpr.options["categorical_kernel"]
            self.mixint = MixedIntegerContext(
                self.design_space,
                work_in_folded_space=True,
            )

            underlying_gpr = self.gpr
            self.gpr = self.mixint.build_kriging_model(self.gpr)

            self.categorical_kernel = underlying_gpr.options["categorical_kernel"]
            self.mixint = MixedIntegerContext(
                self.design_space,
                work_in_folded_space=True,
            )
            self._sampling = self.mixint.build_sampling_method(
                random_state=self.options["random_state"],
            )

        else:
            self.mixint = None
            sampling = MixedIntegerSamplingMethod(
                LHS,
                self.design_space,
                criterion="ese",
                random_state=self.options["random_state"],
            )
            self._sampling = lambda n: sampling(n)

            self.categorical_kernel = None

        # Build DOE
        self._evaluator = self.options["evaluator"]
        xdoe = self.options["xdoe"]
        if xdoe is None:
            self.log("Build initial DOE with LHS")
            n_doe = self.options["n_doe"]
            x_doe = self._sampling(n_doe)
        else:
            self.log("Initial DOE given")
            x_doe = np.atleast_2d(xdoe)

        ydoe = self.options["ydoe"]
        if ydoe is None:
            y_doe = self._evaluator.run(fun, x_doe)
        else:  # to save time if y_doe is already given to EGO
            y_doe = ydoe

        return x_doe, y_doe

    def _train_gpr(self, x_data, y_data):
        self.gpr.set_training_values(x_data, y_data)
        if self.gpr.supports["training_derivatives"]:
            for kx in range(self.gpr.nx):
                self.gpr.set_training_derivatives(
                    x_data, y_data[:, 1 + kx].reshape((y_data.shape[0], 1)), kx
                )
        self.gpr.train()

    def _find_best_point(self, x_data=None, y_data=None, enable_tunneling=False):
        """
        Function that analyse a set of x_data and y_data and give back the
        more interesting point to evaluates according to the selected criterion

        Parameters
        ----------

        x_data: ndarray(n_points, nx)
        y_data: ndarray(n_points, 1)

        Returns
        -------

        ndarray(nx, 1): the next best point to evaluate
        boolean: success flag

        """
        self._train_gpr(x_data, y_data)

        criterion = self.options["criterion"]
        n_start = self.options["n_start"]
        n_max_optim = self.options["n_max_optim"]
        method = "SLSQP"
        options = {"maxiter": 200}
        if self.mixint:
            bounds = self.design_space.get_num_bounds()
            cons = []
            for j in range(len(bounds)):
                lower, upper = bounds[j]
                lo = {"type": "ineq", "fun": lambda x, lb=lower, i=j: x[i] - lb}
                up = {"type": "ineq", "fun": lambda x, ub=upper, i=j: ub - x[i]}
                cons.append(lo)
                cons.append(up)
            bounds = None
            options = {"maxiter": 300}
        else:
            bounds = self.design_space.get_num_bounds()
            cons = ()

        if criterion == "EI":
            self.obj_k = lambda x: -self.EI(np.atleast_2d(x), enable_tunneling, x_data)
        elif criterion == "SBO":
            self.obj_k = lambda x: self.SBO(np.atleast_2d(x))
        elif criterion == "LCB":
            self.obj_k = lambda x: self.LCB(np.atleast_2d(x))

        success = False
        n_optim = 1  # in order to have some success optimizations with SLSQP
        while not success and n_optim <= n_max_optim:
            opt_all = []
            x_start = self._sampling(n_start)
            for ii in range(n_start):
                try:
                    opt_all.append(
                        minimize(
                            lambda x: float(np.array(self.obj_k(x)).flat[0]),
                            x_start[ii, :],
                            method=method,
                            bounds=bounds,
                            constraints=cons,
                            options=options,
                        )
                    )

                except ValueError:  # in case "x0 violates bound constraints" error
                    print("warning: `x0` violates bound constraints")
                    print("x0={}".format(x_start[ii, :]))
                    print("bounds={}".format(bounds))
                    opt_all.append({"success": False})

            opt_all = np.asarray(opt_all)
            for opt_i in opt_all:
                if (
                    opt_i["message"]
                    == "Maximum number of function evaluations has been exceeded."
                ):
                    opt_i["success"] = True
            opt_success = opt_all[[opt_i["success"] for opt_i in opt_all]]
            obj_success = np.array([opt_i["fun"] for opt_i in opt_success])
            success = obj_success.size != 0
            if not success:
                self.log("New start point for the internal optimization")
                n_optim += 1
        if n_optim >= n_max_optim:
            # self.log("Internal optimization failed at EGO iter = {}".format(k))
            return np.atleast_2d(0), False
        ind_min = np.argmin(obj_success)
        opt = opt_success[ind_min]
        x_et_k = np.atleast_2d(opt["x"])

        return x_et_k, True

    def _get_virtual_point(self, x, y_data):
        """
        Depending on the qEI attribute return a predicted value at given point x

        Parameters
        ----------

        x: ndarray(1, 1) the x-coord point where to forecast the y-coord virtual point
        y_data: current y evaluation list only used when qEI is CLmin

        Returns
        -------

        ndarray(1, 1): the so-called virtual y-coord point

        """
        qEI = self.options["qEI"]

        if qEI == "CLmin":
            return np.min(y_data)

        if qEI == "KB":
            return self.gpr.predict_values(x)
        if qEI == "KBUB":
            conf = 3.0

        if qEI == "KBLB":
            conf = -3.0

        if qEI == "KBRand":
            conf = np.random.randn()
        pred = self.gpr.predict_values(x)
        var = self.gpr.predict_variances(x)

        return pred + conf * np.sqrt(var)



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>
Some functions are copied from gaussian_process submodule (Scikit-learn 0.14)
This package is distributed under New BSD license.
"""

import numpy as np
from enum import Enum
from scipy import linalg, optimize
from copy import deepcopy
import warnings

from smt.surrogate_models.surrogate_model import SurrogateModel
from smt.utils.kriging import (
    differences,
    constant,
    linear,
    quadratic,
    pow_exp,
    squar_exp,
    squar_sin_exp,
    abs_exp,
    act_exp,
    cross_distances,
    matern52,
    matern32,
    gower_componentwise_distances,
    componentwise_distance,
    componentwise_distance_PLS,
    compute_X_cont,
    cross_levels,
    compute_X_cross,
    cross_levels_homo_space,
    MixHrcKernelType,
    matrix_data_corr_levels_cat_matrix,
    matrix_data_corr_levels_cat_mod,
    matrix_data_corr_levels_cat_mod_comps,
)
from smt.utils.misc import standardization
from smt.utils.checks import ensure_2d_array, check_support
from scipy.stats import multivariate_normal as m_norm
from smt.sampling_methods import LHS
from smt.utils.design_space import (
    BaseDesignSpace,
    ensure_design_space,
    CategoricalVariable,
)


class MixIntKernelType(Enum):
    EXP_HOMO_HSPHERE = "EXP_HOMO_HSPHERE"
    HOMO_HSPHERE = "HOMO_HSPHERE"
    CONT_RELAX = "CONT_RELAX"
    GOWER = "GOWER"
    COMPOUND_SYMMETRY = "COMPOUND_SYMMETRY"


class KrgBased(SurrogateModel):
    _regression_types = {"constant": constant, "linear": linear, "quadratic": quadratic}

    _correlation_types = {
        "pow_exp": pow_exp,
        "abs_exp": abs_exp,
        "squar_exp": squar_exp,
        "squar_sin_exp": squar_sin_exp,
        "act_exp": act_exp,
        "matern52": matern52,
        "matern32": matern32,
    }

    name = "KrigingBased"

    def _initialize(self):
        super(KrgBased, self)._initialize()
        declare = self.options.declare
        supports = self.supports
        declare(
            "poly",
            "constant",
            values=("constant", "linear", "quadratic"),
            desc="Regression function type",
            types=(str),
        )
        declare(
            "corr",
            "squar_exp",
            values=(
                "pow_exp",
                "abs_exp",
                "squar_exp",
                "act_exp",
                "matern52",
                "matern32",
            ),
            desc="Correlation function type",
        )
        declare(
            "pow_exp_power",
            1.9,
            types=(float),
            desc="Power for the pow_exp kernel function (valid values in (0.0, 2.0]). \
                This option is set automatically when corr option is squar, abs, or matern.",
        )
        declare(
            "categorical_kernel",
            MixIntKernelType.CONT_RELAX,
            values=[
                MixIntKernelType.CONT_RELAX,
                MixIntKernelType.GOWER,
                MixIntKernelType.EXP_HOMO_HSPHERE,
                MixIntKernelType.HOMO_HSPHERE,
                MixIntKernelType.COMPOUND_SYMMETRY,
            ],
            desc="The kernel to use for categorical inputs. Only for non continuous Kriging",
        )
        declare(
            "hierarchical_kernel",
            MixHrcKernelType.ALG_KERNEL,
            values=[
                MixHrcKernelType.ALG_KERNEL,
                MixHrcKernelType.ARC_KERNEL,
            ],
            desc="The kernel to use for mixed hierarchical inputs. Only for non continuous Kriging",
        )
        declare(
            "nugget",
            100.0 * np.finfo(np.double).eps,
            types=(float),
            desc="a jitter for numerical stability",
        )
        declare(
            "theta0", [1e-2], types=(list, np.ndarray), desc="Initial hyperparameters"
        )
        # In practice, in 1D and for X in [0,1], theta^{-2} in [1e-2,infty), i.e.
        # theta in (0,1e1], is a good choice to avoid overfitting. By standardising
        # X in R, X_norm = (X-X_mean)/X_std, then X_norm in [-1,1] if considering
        # one std intervals. This leads to theta in (0,2e1]
        declare(
            "theta_bounds",
            [1e-6, 2e1],
            types=(list, np.ndarray),
            desc="bounds for hyperparameters",
        )
        declare(
            "hyper_opt",
            "TNC",
            values=("Cobyla", "TNC"),
            desc="Optimiser for hyperparameters optimisation",
            types=str,
        )
        declare(
            "eval_noise",
            False,
            types=bool,
            values=(True, False),
            desc="noise evaluation flag",
        )
        declare(
            "noise0",
            [0.0],
            types=(list, np.ndarray),
            desc="Initial noise hyperparameters",
        )
        declare(
            "noise_bounds",
            [100.0 * np.finfo(np.double).eps, 1e10],
            types=(list, np.ndarray),
            desc="bounds for noise hyperparameters",
        )
        declare(
            "use_het_noise",
            False,
            types=bool,
            values=(True, False),
            desc="heteroscedastic noise evaluation flag",
        )
        declare(
            "n_start",
            10,
            types=int,
            desc="number of optimizer runs (multistart method)",
        )
        declare(
            "xlimits",
            None,
            types=(list, np.ndarray),
            desc="definition of a design space of float (continuous) variables: "
            "array-like of size nx x 2 (lower, upper bounds)",
        )
        declare(
            "design_space",
            None,
            types=(BaseDesignSpace, list, np.ndarray),
            desc="definition of the (hierarchical) design space: "
            "use `smt.utils.design_space.DesignSpace` as the main API. Also accepts list of float variable bounds",
        )
        self.options.declare(
            "random_state",
            default=41,
            types=(type(None), int, np.random.RandomState),
            desc="Numpy RandomState object or seed number which controls random draws \
                for internal optim (set by default to get reproductibility)",
        )
        self.best_iteration_fail = None
        self.nb_ill_matrix = 5
        self.is_acting_points = {}

        supports["derivatives"] = True
        supports["variances"] = True
        supports["variance_derivatives"] = True
        supports["x_hierarchy"] = True

    def _final_initialize(self):
        if isinstance(self.options["random_state"], np.random.RandomState):
            self.random_state = self.options["random_state"]
        elif isinstance(self.options["random_state"], int):
            self.random_state = np.random.RandomState(self.options["random_state"])
        else:
            self.random_state = np.random.RandomState()

        # initialize default power values
        if self.options["corr"] == "squar_exp":
            self.options["pow_exp_power"] = 2.0
        elif self.options["corr"] in [
            "abs_exp",
            "squar_sin_exp",
            "matern32",
            "matern52",
        ]:
            self.options["pow_exp_power"] = 1.0

        # Check the pow_exp_power is >0 and <=2
        assert (
            self.options["pow_exp_power"] > 0 and self.options["pow_exp_power"] <= 2
        ), (
            "The power value for exponential power function can only be >0 and <=2, but %s was given"
            % self.options["pow_exp_power"]
        )

    @property
    def design_space(self) -> BaseDesignSpace:
        xt = self.training_points.get(None)
        if xt is not None:
            xt = xt[0][0]

        if self.options["design_space"] is None:
            self.options["design_space"] = ensure_design_space(
                xt=xt, xlimits=self.options["xlimits"]
            )

        elif not isinstance(self.options["design_space"], BaseDesignSpace):
            ds_input = self.options["design_space"]
            self.options["design_space"] = ensure_design_space(
                xt=xt, xlimits=ds_input, design_space=ds_input
            )
        return self.options["design_space"]

    @property
    def is_continuous(self) -> bool:
        return self.design_space.is_all_cont

    def set_training_values(
        self, xt: np.ndarray, yt: np.ndarray, name=None, is_acting=None
    ) -> None:
        """
        Set training data (values).

        Parameters
        ----------
        xt : np.ndarray[nt, nx] or np.ndarray[nt]
            The input values for the nt training points.
        yt : np.ndarray[nt, ny] or np.ndarray[nt]
            The output values for the nt training points.
        name : str or None
            An optional label for the group of training points being set.
            This is only used in special situations (e.g., multi-fidelity applications).
        is_acting : np.ndarray[nt, nx] or np.ndarray[nt]
            Matrix specifying which of the design variables is acting in a hierarchical design space
        """
        super().set_training_values(xt, yt, name=name)
        if is_acting is not None:
            self.is_acting_points[name] = is_acting

    def _correct_distances_cat_decreed(
        self,
        D,
        is_acting,
        listcatdecreed,
        ij,
        is_acting_y=None,
        mixint_type=MixIntKernelType.CONT_RELAX,
    ):
        indjcat = -1
        for j in listcatdecreed:
            indjcat = indjcat + 1
            if j:
                indicat = -1
                indices = 0
                for v in range(len(self.design_space.design_variables)):
                    if isinstance(
                        self.design_space.design_variables[v], CategoricalVariable
                    ):
                        indicat = indicat + 1
                        if indicat == indjcat:
                            ia2 = np.zeros((len(ij), 2), dtype=bool)
                            if is_acting_y is None:
                                ia2 = (is_acting[:, self.cat_features][:, indjcat])[ij]
                            else:
                                ia2[:, 0] = (
                                    is_acting[:, self.cat_features][:, indjcat]
                                )[ij[:, 0]]
                                ia2[:, 1] = (
                                    is_acting_y[:, self.cat_features][:, indjcat]
                                )[ij[:, 1]]

                            act_inact = ia2[:, 0] ^ ia2[:, 1]
                            act_act = ia2[:, 0] & ia2[:, 1]

                            if mixint_type == MixIntKernelType.CONT_RELAX:
                                val_act = (
                                    np.array([1] * self.n_levels[indjcat])
                                    - self.X2_offset[
                                        indices : indices + self.n_levels[indjcat]
                                    ]
                                ) / self.X2_scale[
                                    indices : indices + self.n_levels[indjcat]
                                ] - (
                                    np.array([0] * self.n_levels[indjcat])
                                    - self.X2_offset[
                                        indices : indices + self.n_levels[indjcat]
                                    ]
                                ) / self.X2_scale[
                                    indices : indices + self.n_levels[indjcat]
                                ]
                                D[:, indices : indices + self.n_levels[indjcat]][
                                    act_inact
                                ] = val_act
                                D[:, indices : indices + self.n_levels[indjcat]][
                                    act_act
                                ] = (
                                    np.sqrt(2)
                                    * D[:, indices : indices + self.n_levels[indjcat]][
                                        act_act
                                    ]
                                )
                            elif mixint_type == MixIntKernelType.GOWER:
                                D[:, indices : indices + 1][act_inact] = (
                                    self.n_levels[indjcat] * 0.5
                                )
                                D[:, indices : indices + 1][act_act] = (
                                    np.sqrt(2) * D[:, indices : indices + 1][act_act]
                                )

                            else:
                                raise ValueError(
                                    "Continuous decreed kernel not implemented"
                                )
                        else:
                            if mixint_type == MixIntKernelType.CONT_RELAX:
                                indices = indices + self.n_levels[indicat]
                            elif mixint_type == MixIntKernelType.GOWER:
                                indices = indices + 1
                    else:
                        indices = indices + 1
        return D

    def _new_train(self):
        # Sampling points X and y
        X = self.training_points[None][0][0]
        y = self.training_points[None][0][1]
        # Get is_acting status from design space model if needed (might correct training points)
        is_acting = self.is_acting_points.get(None)
        if is_acting is None and not self.is_continuous:
            X, is_acting = self.design_space.correct_get_acting(X)
            self.training_points[None][0][0] = X
            self.is_acting_points[None] = is_acting

        # Compute PLS-coefficients (attr of self) and modified X and y (if GEKPLS is used)
        if self.name not in ["Kriging", "MGP", "SGP"]:
            if self.is_continuous:
                X, y = self._compute_pls(X.copy(), y.copy())

        self._check_param()
        self.X_train = X
        self.is_acting_train = is_acting
        self._corr_params = None
        _, self.cat_features = compute_X_cont(self.X_train, self.design_space)
        D = None  # For SGP, D is not computed at all
        # Center and scale X and y
        (
            self.X_norma,
            self.y_norma,
            self.X_offset,
            self.y_mean,
            self.X_scale,
            self.y_std,
        ) = standardization(X.copy(), y.copy())

        if not self.options["eval_noise"]:
            self.optimal_noise = np.array(self.options["noise0"])
        elif self.options["use_het_noise"]:
            # hetGP works with unique design variables when noise variance are not given
            (
                self.X_norma,
                index_unique,
                nt_reps,
            ) = np.unique(self.X_norma, return_inverse=True, return_counts=True, axis=0)
            self.nt = self.X_norma.shape[0]

            # computing the mean of the output per unique design variable (see Binois et al., 2018)
            y_norma_unique = []
            for i in range(self.nt):
                y_norma_unique.append(np.mean(self.y_norma[index_unique == i]))
            # pointwise sensible estimates of the noise variances (see Ankenman et al., 2010)
            self.optimal_noise = self.options["noise0"] * np.ones(self.nt)
            for i in range(self.nt):
                diff = self.y_norma[index_unique == i] - y_norma_unique[i]
                if np.sum(diff**2) != 0.0:
                    self.optimal_noise[i] = np.std(diff, ddof=1) ** 2
            self.optimal_noise = self.optimal_noise / nt_reps
            self.y_norma = y_norma_unique

        if not (self.is_continuous):
            D, self.ij, X_cont = gower_componentwise_distances(
                X=X,
                x_is_acting=is_acting,
                design_space=self.design_space,
                hierarchical_kernel=self.options["hierarchical_kernel"],
            )
            self.Lij, self.n_levels = cross_levels(
                X=self.X_train, ij=self.ij, design_space=self.design_space
            )
            listcatdecreed = self.design_space.is_conditionally_acting[
                self.cat_features
            ]
            if np.any(listcatdecreed):
                D = self._correct_distances_cat_decreed(
                    D,
                    is_acting,
                    listcatdecreed,
                    self.ij,
                    mixint_type=MixIntKernelType.GOWER,
                )
            if self.options["categorical_kernel"] == MixIntKernelType.CONT_RELAX:
                X2, _ = self.design_space.unfold_x(X)
                (
                    self.X2_norma,
                    _,
                    self.X2_offset,
                    _,
                    self.X2_scale,
                    _,
                ) = standardization(X2.copy(), y.copy())
                D, _ = cross_distances(self.X2_norma)
                self.Lij, self.n_levels = cross_levels(
                    X=self.X_train, ij=self.ij, design_space=self.design_space
                )
                listcatdecreed = self.design_space.is_conditionally_acting[
                    self.cat_features
                ]
                if np.any(listcatdecreed):
                    D = self._correct_distances_cat_decreed(
                        D,
                        is_acting,
                        listcatdecreed,
                        self.ij,
                        mixint_type=MixIntKernelType.CONT_RELAX,
                    )

            # Center and scale X_cont and y
            (
                self.X_norma,
                self.y_norma,
                self.X_offset,
                self.y_mean,
                self.X_scale,
                self.y_std,
            ) = standardization(X_cont.copy(), y.copy())

        if self.name not in ["SGP"]:
            if self.is_continuous:
                # Calculate matrix of distances D between samples
                D, self.ij = cross_distances(self.X_norma)

            if np.min(np.sum(np.abs(D), axis=1)) == 0.0:
                warnings.warn(
                    "Warning: multiple x input features have the same value (at least same row twice)."
                )

        ####
        # Regression matrix and parameters
        self.F = self._regression_types[self.options["poly"]](self.X_norma)
        n_samples_F = self.F.shape[0]
        if self.F.ndim > 1:
            p = self.F.shape[1]
        else:
            p = 1
        self._check_F(n_samples_F, p)

        # Optimization
        (
            self.optimal_rlf_value,
            self.optimal_par,
            self.optimal_theta,
        ) = self._optimize_hyperparam(D)
        if self.name in ["MGP"]:
            self._specific_train()
        elif self.name in ["SGP"] and not self.options["use_het_noise"]:
            if self.options["eval_noise"]:
                self.optimal_noise = self.optimal_theta[-1]
                self.optimal_sigma2 = self.optimal_theta[-2]
                self.optimal_theta = self.optimal_theta[:-2]
            else:
                self.optimal_sigma2 = self.optimal_theta[-1]
                self.optimal_theta = self.optimal_theta[:-1]
        else:
            if self.options["eval_noise"] and not self.options["use_het_noise"]:
                self.optimal_noise = self.optimal_theta[-1]
                self.optimal_theta = self.optimal_theta[:-1]
        # if self.name != "MGP":
        #     del self.y_norma, self.D

    def _train(self):
        """
        Train the model
        """
        # outputs['sol'] = self.sol

        self._new_train()

    def _initialize_theta(self, theta, n_levels, cat_features, cat_kernel):
        self.n_levels_origin = n_levels
        if self._corr_params is not None:
            return self._corr_params
        nx = self.nx
        try:
            cat_kernel_comps = self.options["cat_kernel_comps"]
            if cat_kernel_comps is not None:
                n_levels = np.array(cat_kernel_comps)
        except KeyError:
            cat_kernel_comps = None
        try:
            ncomp = self.options["n_comp"]
            try:
                self.pls_coeff_cont
            except AttributeError:
                self.pls_coeff_cont = []
        except KeyError:
            cat_kernel_comps = None
            ncomp = 1e5

        theta_cont_features = np.zeros((len(theta), 1), dtype=bool)
        theta_cat_features = np.ones((len(theta), len(n_levels)), dtype=bool)
        if cat_kernel in [
            MixIntKernelType.EXP_HOMO_HSPHERE,
            MixIntKernelType.HOMO_HSPHERE,
        ]:
            theta_cat_features = np.zeros((len(theta), len(n_levels)), dtype=bool)
        i = 0
        j = 0
        n_theta_cont = 0
        for feat in cat_features:
            if feat:
                if cat_kernel in [
                    MixIntKernelType.EXP_HOMO_HSPHERE,
                    MixIntKernelType.HOMO_HSPHERE,
                ]:
                    theta_cat_features[
                        j : j + int(n_levels[i] * (n_levels[i] - 1) / 2), i
                    ] = [True] * int(n_levels[i] * (n_levels[i] - 1) / 2)
                    j += int(n_levels[i] * (n_levels[i] - 1) / 2)
                i += 1
            else:
                if n_theta_cont < ncomp:
                    theta_cont_features[j] = True
                    theta_cat_features[j] = False
                    j += 1
                    n_theta_cont += 1

        theta_cat_features = (
            [
                np.where(theta_cat_features[:, i_lvl])[0]
                for i_lvl in range(len(n_levels))
            ],
            np.any(theta_cat_features, axis=1) if len(n_levels) > 0 else None,
        )

        self._corr_params = params = (
            cat_kernel_comps,
            ncomp,
            theta_cat_features,
            theta_cont_features,
            nx,
            n_levels,
        )
        return params

    def _matrix_data_corr(
        self,
        corr,
        design_space,
        power,
        theta,
        theta_bounds,
        dx,
        Lij,
        n_levels,
        cat_features,
        cat_kernel,
        x=None,
    ):
        """
        matrix kernel correlation model.

        Parameters
        ----------
        corr: correlation_types
            - The autocorrelation model
        design_space: BaseDesignSpace
            - The design space definition
        theta : list[small_d * n_comp]
            Hyperparameters of the correlation model
        dx: np.ndarray[n_obs * (n_obs - 1) / 2, n_comp]
            - The gower_componentwise_distances between the samples.
        Lij: np.ndarray [n_obs * (n_obs - 1) / 2, 2]
                - The levels corresponding to the indices i and j of the vectors in X.
        n_levels: np.ndarray
                - The number of levels for every categorical variable.
        cat_features: np.ndarray [dim]
            -  Indices of the categorical input dimensions.
         cat_kernel : string
             - The kernel to use for categorical inputs. Only for non continuous Kriging",
        x : np.ndarray[n_obs , n_comp]
            - The input instead of dx for homo_hs prediction
        Returns
        -------
        r: np.ndarray[n_obs * (n_obs - 1) / 2,1]
            An array containing the values of the autocorrelation model.
        """

        _correlation_types = {
            "pow_exp": pow_exp,
            "abs_exp": abs_exp,
            "squar_exp": squar_exp,
            "squar_sin_exp": squar_sin_exp,
            "act_exp": act_exp,
            "matern52": matern52,
            "matern32": matern32,
        }

        # Initialize static parameters
        (
            cat_kernel_comps,
            ncomp,
            theta_cat_features,
            theta_cont_features,
            nx,
            n_levels,
        ) = self._initialize_theta(theta, n_levels, cat_features, cat_kernel)

        # Sampling points X and y
        if "MFK" in self.name:
            if self._lvl < self.nlvl - 1:
                X = self.training_points[self._lvl][0][0]
                y = self.training_points[self._lvl][0][1]
            elif self._lvl == self.nlvl - 1:
                X = self.training_points[None][0][0]
                y = self.training_points[None][0][1]
        else:
            X = self.training_points[None][0][0]
            y = self.training_points[None][0][1]

        if cat_kernel == MixIntKernelType.CONT_RELAX:
            X_pls_space, _ = design_space.unfold_x(X)
            nx = len(theta)

        elif cat_kernel == MixIntKernelType.GOWER:
            X_pls_space = np.copy(X)
        else:
            X_pls_space, _ = compute_X_cont(X, design_space)
        if cat_kernel_comps is not None or ncomp < 1e5:
            if np.size(self.pls_coeff_cont) == 0:
                X, y = self._compute_pls(X_pls_space.copy(), y.copy())
                self.pls_coeff_cont = self.coeff_pls
            if cat_kernel in [MixIntKernelType.GOWER, MixIntKernelType.CONT_RELAX]:
                d = componentwise_distance_PLS(
                    dx,
                    corr,
                    self.options["n_comp"],
                    self.pls_coeff_cont,
                    power,
                    theta=None,
                    return_derivative=False,
                )
                r = _correlation_types[corr](theta, d)
                return r
            else:
                d_cont = componentwise_distance_PLS(
                    dx[:, np.logical_not(cat_features)],
                    corr,
                    self.options["n_comp"],
                    self.pls_coeff_cont,
                    power,
                    theta=None,
                    return_derivative=False,
                )
        else:
            d = componentwise_distance(
                dx,
                corr,
                nx,
                power,
                theta=None,
                return_derivative=False,
            )
            if cat_kernel in [MixIntKernelType.GOWER, MixIntKernelType.CONT_RELAX]:
                r = _correlation_types[corr](theta, d)
                return r
            else:
                d_cont = d[:, np.logical_not(cat_features)]
        if self.options["corr"] == "squar_sin_exp":
            if self.options["categorical_kernel"] != MixIntKernelType.GOWER:
                theta_cont_features[-len([self.design_space.is_cat_mask]) :] = (
                    np.atleast_2d(
                        np.array([True] * len([self.design_space.is_cat_mask]))
                    ).T
                )
                theta_cat_features[1][-len([self.design_space.is_cat_mask]) :] = (
                    np.atleast_2d(
                        np.array([False] * len([self.design_space.is_cat_mask]))
                    ).T
                )

        theta_cont = theta[theta_cont_features[:, 0]]
        r_cont = _correlation_types[corr](theta_cont, d_cont)
        r_cat = np.copy(r_cont) * 0
        r = np.copy(r_cont)
        ##Theta_cat_i loop
        try:
            self.coeff_pls_cat
        except AttributeError:
            self.coeff_pls_cat = []

        theta_cat_kernel = theta
        if len(n_levels) > 0:
            theta_cat_kernel = theta.copy()
            if cat_kernel == MixIntKernelType.EXP_HOMO_HSPHERE:
                theta_cat_kernel[theta_cat_features[1]] *= 0.5 * np.pi / theta_bounds[1]
            elif cat_kernel == MixIntKernelType.HOMO_HSPHERE:
                theta_cat_kernel[theta_cat_features[1]] *= 2.0 * np.pi / theta_bounds[1]
            elif cat_kernel == MixIntKernelType.COMPOUND_SYMMETRY:
                theta_cat_kernel[theta_cat_features[1]] *= 2.0
                theta_cat_kernel[theta_cat_features[1]] -= (
                    theta_bounds[1] + theta_bounds[0]
                )
                theta_cat_kernel[theta_cat_features[1]] *= 1 / (
                    1.000000000001 * theta_bounds[1]
                )

        for i in range(len(n_levels)):
            theta_cat = theta_cat_kernel[theta_cat_features[0][i]]
            if cat_kernel == MixIntKernelType.COMPOUND_SYMMETRY:
                T = np.zeros((n_levels[i], n_levels[i]))
                for tij in range(n_levels[i]):
                    for tji in range(n_levels[i]):
                        if tij == tji:
                            T[tij, tji] = 1
                        else:
                            T[tij, tji] = max(
                                theta_cat[0], 1e-10 - 1 / (n_levels[i] - 1)
                            )
            else:
                T = matrix_data_corr_levels_cat_matrix(
                    i,
                    n_levels,
                    theta_cat,
                    theta_bounds,
                    is_ehh=cat_kernel == MixIntKernelType.EXP_HOMO_HSPHERE,
                )

            if cat_kernel_comps is not None:
                # Sampling points X and y
                X = self.training_points[None][0][0]
                y = self.training_points[None][0][1]
                X_icat = X[:, cat_features]
                X_icat = X_icat[:, i]
                old_n_comp = (
                    self.options["n_comp"] if "n_comp" in self.options else None
                )
                self.options["n_comp"] = int(n_levels[i] / 2 * (n_levels[i] - 1))
                X_full_space = compute_X_cross(X_icat, self.n_levels_origin[i])
                try:
                    self.coeff_pls = self.coeff_pls_cat[i]
                except IndexError:
                    _, _ = self._compute_pls(X_full_space.copy(), y.copy())
                    self.coeff_pls_cat.append(self.coeff_pls)

                if x is not None:
                    x_icat = x[:, cat_features]
                    x_icat = x_icat[:, i]
                    x_full_space = compute_X_cross(x_icat, self.n_levels_origin[i])
                    dx_cat_i = cross_levels_homo_space(
                        x_full_space, self.ij, y=X_full_space
                    )
                else:
                    dx_cat_i = cross_levels_homo_space(X_full_space, self.ij)

                d_cat_i = componentwise_distance_PLS(
                    dx_cat_i,
                    "squar_exp",
                    self.options["n_comp"],
                    self.coeff_pls,
                    power=self.options["pow_exp_power"],
                    theta=None,
                    return_derivative=False,
                )

                matrix_data_corr_levels_cat_mod_comps(
                    i,
                    Lij,
                    r_cat,
                    n_levels,
                    T,
                    d_cat_i,
                    has_cat_kernel=cat_kernel
                    in [
                        MixIntKernelType.EXP_HOMO_HSPHERE,
                        MixIntKernelType.HOMO_HSPHERE,
                    ],
                )
            else:
                matrix_data_corr_levels_cat_mod(
                    i,
                    Lij,
                    r_cat,
                    T,
                    has_cat_kernel=cat_kernel
                    in [
                        MixIntKernelType.EXP_HOMO_HSPHERE,
                        MixIntKernelType.HOMO_HSPHERE,
                        MixIntKernelType.COMPOUND_SYMMETRY,
                    ],
                )

            r = np.multiply(r, r_cat)
            if cat_kernel_comps is not None:
                if old_n_comp is None:
                    self.options._dict.pop("n_comp", None)
                else:
                    self.options["n_comp"] = old_n_comp
        return r

    def _reduced_likelihood_function(self, theta):
        """
        This function determines the BLUP parameters and evaluates the reduced
        likelihood function for the given autocorrelation parameters theta.
        Maximizing this function wrt the autocorrelation parameters theta is
        equivalent to maximizing the likelihood of the assumed joint Gaussian
        distribution of the observations y evaluated onto the design of
        experiments X.

        Parameters
        ----------
        theta: list(n_comp), optional
            - An array containing the autocorrelation parameters at which the
              Gaussian Process model parameters should be determined.

        Returns
        -------
        reduced_likelihood_function_value: real
            - The value of the reduced likelihood function associated to the
              given autocorrelation parameters theta.
        par: dict()
            - A dictionary containing the requested Gaussian Process model
              parameters:
            sigma2
            Gaussian Process variance.
            beta
            Generalized least-squares regression weights for
            Universal Kriging or for Ordinary Kriging.
            gamma
            Gaussian Process weights.
            C
            Cholesky decomposition of the correlation matrix [R].
            Ft
            Solution of the linear equation system : [R] x Ft = F
            Q, G
            QR decomposition of the matrix Ft.
        """
        # Initialize output

        reduced_likelihood_function_value = -np.inf
        par = {}
        # Set up R
        nugget = self.options["nugget"]
        if self.options["eval_noise"]:
            nugget = 0

        noise = self.noise0
        tmp_var = theta
        if self.options["use_het_noise"]:
            noise = self.optimal_noise
        if self.options["eval_noise"] and not self.options["use_het_noise"]:
            theta = tmp_var[0:-1]
            noise = tmp_var[-1]
        if not (self.is_continuous):
            dx = self.D
            if self.options["categorical_kernel"] == MixIntKernelType.CONT_RELAX:
                if "MFK" in self.name:
                    if (
                        self._lvl == self.nlvl - 1
                    ):  # highest fidelity identified by the key None
                        X2, _ = self.design_space.unfold_x(
                            self.training_points[None][0][0]
                        )
                        self.X2_norma[str(self._lvl)] = (
                            X2 - self.X2_offset
                        ) / self.X2_scale
                        dx, _ = cross_distances(self.X2_norma[str(self._lvl)])
                    elif self._lvl < self.nlvl - 1:
                        X2, _ = self.design_space.unfold_x(
                            self.training_points[self._lvl][0][0]
                        )
                        self.X2_norma[str(self._lvl)] = (
                            X2 - self.X2_offset
                        ) / self.X2_scale
                        dx, _ = cross_distances(self.X2_norma[str(self._lvl)])
                else:
                    X2, _ = self.design_space.unfold_x(self.training_points[None][0][0])
                    (
                        self.X2_norma,
                        _,
                        self.X2_offset,
                        _,
                        self.X2_scale,
                        _,
                    ) = standardization(X2, self.training_points[None][0][1])
                    dx, _ = cross_distances(self.X2_norma)

            r = self._matrix_data_corr(
                corr=self.options["corr"],
                design_space=self.design_space,
                power=self.options["pow_exp_power"],
                theta=theta,
                theta_bounds=self.options["theta_bounds"],
                dx=dx,
                Lij=self.Lij,
                n_levels=self.n_levels,
                cat_features=self.cat_features,
                cat_kernel=self.options["categorical_kernel"],
            ).reshape(-1, 1)
        else:
            r = self._correlation_types[self.options["corr"]](theta, self.D).reshape(
                -1, 1
            )
        R = np.eye(self.nt) * (1.0 + nugget + noise)
        R[self.ij[:, 0], self.ij[:, 1]] = r[:, 0]
        R[self.ij[:, 1], self.ij[:, 0]] = r[:, 0]
        # Cholesky decomposition of R

        try:
            C = linalg.cholesky(R, lower=True)
        except (linalg.LinAlgError, ValueError) as e:
            print("exception : ", e)
            print(np.linalg.eig(R)[0])
            return reduced_likelihood_function_value, par
        if linalg.svd(R, compute_uv=False)[-1] < 1.1 * nugget:
            warnings.warn(
                "R is too ill conditioned. Poor combination "
                "of regression model and observations."
            )

        # Get generalized least squared solution
        Ft = linalg.solve_triangular(C, self.F, lower=True)
        Q, G = linalg.qr(Ft, mode="economic")
        sv = linalg.svd(G, compute_uv=False)
        rcondG = sv[-1] / sv[0]
        if rcondG < 1e-10:
            # Check F
            sv = linalg.svd(self.F, compute_uv=False)
            condF = sv[0] / sv[-1]
            if condF > 1e15:
                raise Exception(
                    "F is too ill conditioned. Poor combination "
                    "of regression model and observations."
                )

            else:
                # Ft is too ill conditioned, get out (try different theta)
                return reduced_likelihood_function_value, par

        Yt = linalg.solve_triangular(C, self.y_norma, lower=True)
        beta = linalg.solve_triangular(G, np.dot(Q.T, Yt))
        rho = Yt - np.dot(Ft, beta)

        # The determinant of R is equal to the squared product of the diagonal
        # elements of its Cholesky decomposition C
        detR = (np.diag(C) ** (2.0 / self.nt)).prod()
        # Compute/Organize output
        p = 0
        q = 0
        if self.name in ["MFK", "MFKPLS", "MFKPLSK"]:
            p = self.p
            q = self.q
        sigma2 = (rho**2.0).sum(axis=0) / (self.nt - p - q)
        reduced_likelihood_function_value = -(self.nt - p - q) * np.log10(
            sigma2.sum()
        ) - self.nt * np.log10(detR)
        par["sigma2"] = sigma2 * self.y_std**2.0
        par["beta"] = beta
        par["gamma"] = linalg.solve_triangular(C.T, rho)
        par["C"] = C
        par["Ft"] = Ft
        par["G"] = G
        par["Q"] = Q

        if self.name in ["MGP"]:
            reduced_likelihood_function_value += self._reduced_log_prior(theta)

        # A particular case when f_min_cobyla fail
        if (self.best_iteration_fail is not None) and (
            not np.isinf(reduced_likelihood_function_value)
        ):
            if reduced_likelihood_function_value > self.best_iteration_fail:
                self.best_iteration_fail = reduced_likelihood_function_value
                self._thetaMemory = np.array(tmp_var)

        elif (self.best_iteration_fail is None) and (
            not np.isinf(reduced_likelihood_function_value)
        ):
            self.best_iteration_fail = reduced_likelihood_function_value
            self._thetaMemory = np.array(tmp_var)
        if reduced_likelihood_function_value > 1e15:
            reduced_likelihood_function_value = 1e15
        return reduced_likelihood_function_value, par

    def _reduced_likelihood_gradient(self, theta):
        """
        Evaluates the reduced_likelihood_gradient at a set of hyperparameters.

        Parameters
        ---------
        theta : list(n_comp), optional
            - An array containing the autocorrelation parameters at which the
              Gaussian Process model parameters should be determined.

        Returns
        -------
        grad_red : np.ndarray (dim,1)
            Derivative of the reduced_likelihood
        par: dict()
            - A dictionary containing the requested Gaussian Process model
              parameters:
            sigma2
            Gaussian Process variance.
            beta
            Generalized least-squares regression weights for
            Universal Kriging or for Ordinary Kriging.
            gamma
            Gaussian Process weights.
            C
            Cholesky decomposition of the correlation matrix [R].
            Ft
            Solution of the linear equation system : [R] x Ft = F
            Q, G
            QR decomposition of the matrix Ft.
            dr
            List of all the correlation matrix derivative
            tr
            List of all the trace part in the reduce likelihood derivatives
            dmu
            List of all the mean derivatives
            arg
            List of all minus_Cinv_dRdomega_gamma
            dsigma
            List of all sigma derivatives
        """
        red, par = self._reduced_likelihood_function(theta)

        C = par["C"]
        gamma = par["gamma"]
        Q = par["Q"]
        G = par["G"]
        sigma_2 = par["sigma2"] + self.options["nugget"]

        nb_theta = len(theta)
        grad_red = np.zeros(nb_theta)

        dr_all = []
        tr_all = []
        dmu_all = []
        arg_all = []
        dsigma_all = []
        dbeta_all = []
        for i_der in range(nb_theta):
            # Compute R derivatives
            dr = self._correlation_types[self.options["corr"]](
                theta, self.D, grad_ind=i_der
            )

            dr_all.append(dr)

            dR = np.zeros((self.nt, self.nt))
            dR[self.ij[:, 0], self.ij[:, 1]] = dr[:, 0]
            dR[self.ij[:, 1], self.ij[:, 0]] = dr[:, 0]

            # Compute beta derivatives
            Cinv_dR_gamma = linalg.solve_triangular(C, np.dot(dR, gamma), lower=True)
            dbeta = -linalg.solve_triangular(G, np.dot(Q.T, Cinv_dR_gamma))
            arg_all.append(Cinv_dR_gamma)

            dbeta_all.append(dbeta)

            # Compute mu derivatives
            dmu = np.dot(self.F, dbeta)
            dmu_all.append(dmu)

            # Compute log(detR) derivatives
            tr_1 = linalg.solve_triangular(C, dR, lower=True)
            tr = linalg.solve_triangular(C.T, tr_1)
            tr_all.append(tr)

            # Compute Sigma2 Derivatives
            dsigma_2 = (
                (1 / self.nt)
                * (
                    -dmu.T.dot(gamma)
                    - gamma.T.dot(dmu)
                    - np.dot(gamma.T, dR.dot(gamma))
                )
                * self.y_std**2.0
            )
            dsigma_all.append(dsigma_2)

            # Compute reduced log likelihood derivatives
            grad_red[i_der] = (
                -self.nt / np.log(10) * (dsigma_2 / sigma_2 + np.trace(tr) / self.nt)
            ).item()

        par["dr"] = dr_all
        par["tr"] = tr_all
        par["dmu"] = dmu_all
        par["arg"] = arg_all
        par["dsigma"] = dsigma_all
        par["dbeta_all"] = dbeta_all

        grad_red = np.atleast_2d(grad_red).T

        if self.name in ["MGP"]:
            grad_red += self._reduced_log_prior(theta, grad=True)
        return grad_red, par

    def _reduced_likelihood_hessian(self, theta):
        """
        Evaluates the reduced_likelihood_gradient at a set of hyperparameters.

        Parameters
        ----------
        theta : list(n_comp), optional
            - An array containing the autocorrelation parameters at which the
              Gaussian Process model parameters should be determined.

        Returns
        -------
        hess : np.ndarray
            Hessian values.
        hess_ij: np.ndarray [nb_theta * (nb_theta + 1) / 2, 2]
            - The indices i and j of the vectors in theta associated to the hessian in hess.
        par: dict()
            - A dictionary containing the requested Gaussian Process model
              parameters:
            sigma2
            Gaussian Process variance.
            beta
            Generalized least-squared regression weights for
            Universal Kriging or for Ordinary Kriging.
            gamma
            Gaussian Process weights.
            C
            Cholesky decomposition of the correlation matrix [R].
            Ft
            Solution of the linear equation system : [R] x Ft = F
            Q, G
            QR decomposition of the matrix Ft.
            dr
            List of all the correlation matrix derivative
            tr
            List of all the trace part in the reduce likelihood derivatives
            dmu
            List of all the mean derivatives
            arg
            List of all minus_Cinv_dRdomega_gamma
            dsigma
            List of all sigma derivatives
        """
        dred, par = self._reduced_likelihood_gradient(theta)

        C = par["C"]
        gamma = par["gamma"]
        Q = par["Q"]
        G = par["G"]
        sigma_2 = par["sigma2"]

        nb_theta = len(theta)

        dr_all = par["dr"]
        tr_all = par["tr"]
        dmu_all = par["dmu"]
        arg_all = par["arg"]
        dsigma = par["dsigma"]
        Rinv_dRdomega_gamma_all = []
        Rinv_dmudomega_all = []

        n_val_hess = nb_theta * (nb_theta + 1) // 2
        hess_ij = np.zeros((n_val_hess, 2), dtype=np.int32)
        hess = np.zeros((n_val_hess, 1))
        ind_1 = 0
        if self.name in ["MGP"]:
            log_prior = self._reduced_log_prior(theta, hessian=True)

        for omega in range(nb_theta):
            ind_0 = ind_1
            ind_1 = ind_0 + nb_theta - omega
            hess_ij[ind_0:ind_1, 0] = omega
            hess_ij[ind_0:ind_1, 1] = np.arange(omega, nb_theta)

            dRdomega = np.zeros((self.nt, self.nt))
            dRdomega[self.ij[:, 0], self.ij[:, 1]] = dr_all[omega][:, 0]
            dRdomega[self.ij[:, 1], self.ij[:, 0]] = dr_all[omega][:, 0]

            dmudomega = dmu_all[omega]
            Cinv_dmudomega = linalg.solve_triangular(C, dmudomega, lower=True)
            Rinv_dmudomega = linalg.solve_triangular(C.T, Cinv_dmudomega)
            Rinv_dmudomega_all.append(Rinv_dmudomega)
            Rinv_dRdomega_gamma = linalg.solve_triangular(C.T, arg_all[omega])
            Rinv_dRdomega_gamma_all.append(Rinv_dRdomega_gamma)

            for i, eta in enumerate(hess_ij[ind_0:ind_1, 1]):
                dRdeta = np.zeros((self.nt, self.nt))
                dRdeta[self.ij[:, 0], self.ij[:, 1]] = dr_all[eta][:, 0]
                dRdeta[self.ij[:, 1], self.ij[:, 0]] = dr_all[eta][:, 0]

                dr_eta_omega = self._correlation_types[self.options["corr"]](
                    theta, self.D, grad_ind=omega, hess_ind=eta
                )
                dRdetadomega = np.zeros((self.nt, self.nt))
                dRdetadomega[self.ij[:, 0], self.ij[:, 1]] = dr_eta_omega[:, 0]
                dRdetadomega[self.ij[:, 1], self.ij[:, 0]] = dr_eta_omega[:, 0]

                # Compute beta second derivatives
                dRdeta_Rinv_dmudomega = np.dot(dRdeta, Rinv_dmudomega)

                dmudeta = dmu_all[eta]
                Cinv_dmudeta = linalg.solve_triangular(C, dmudeta, lower=True)
                Rinv_dmudeta = linalg.solve_triangular(C.T, Cinv_dmudeta)
                dRdomega_Rinv_dmudeta = np.dot(dRdomega, Rinv_dmudeta)

                dRdeta_Rinv_dRdomega_gamma = np.dot(dRdeta, Rinv_dRdomega_gamma)

                Rinv_dRdeta_gamma = linalg.solve_triangular(C.T, arg_all[eta])
                dRdomega_Rinv_dRdeta_gamma = np.dot(dRdomega, Rinv_dRdeta_gamma)

                dRdetadomega_gamma = np.dot(dRdetadomega, gamma)

                beta_sum = (
                    dRdeta_Rinv_dmudomega
                    + dRdomega_Rinv_dmudeta
                    + dRdeta_Rinv_dRdomega_gamma
                    + dRdomega_Rinv_dRdeta_gamma
                    - dRdetadomega_gamma
                )

                Qt_Cinv_beta_sum = np.dot(
                    Q.T, linalg.solve_triangular(C, beta_sum, lower=True)
                )
                dbetadetadomega = linalg.solve_triangular(G, Qt_Cinv_beta_sum)

                # Compute mu second derivatives
                dmudetadomega = np.dot(self.F, dbetadetadomega)

                # Compute sigma2 second derivatives
                sigma_arg_1 = (
                    -np.dot(dmudetadomega.T, gamma)
                    + np.dot(dmudomega.T, Rinv_dRdeta_gamma)
                    + np.dot(dmudeta.T, Rinv_dRdomega_gamma)
                )

                sigma_arg_2 = (
                    -np.dot(gamma.T, dmudetadomega)
                    + np.dot(gamma.T, dRdeta_Rinv_dmudomega)
                    + np.dot(gamma.T, dRdomega_Rinv_dmudeta)
                )

                sigma_arg_3 = np.dot(dmudeta.T, Rinv_dmudomega) + np.dot(
                    dmudomega.T, Rinv_dmudeta
                )

                sigma_arg_4_in = (
                    -dRdetadomega_gamma
                    + dRdeta_Rinv_dRdomega_gamma
                    + dRdomega_Rinv_dRdeta_gamma
                )
                sigma_arg_4 = np.dot(gamma.T, sigma_arg_4_in)

                dsigma2detadomega = (
                    (1 / self.nt)
                    * (sigma_arg_1 + sigma_arg_2 + sigma_arg_3 + sigma_arg_4)
                    * self.y_std**2.0
                )

                # Compute Hessian
                dreddetadomega_tr_1 = np.trace(np.dot(tr_all[eta], tr_all[omega]))

                dreddetadomega_tr_2 = np.trace(
                    linalg.solve_triangular(
                        C.T, linalg.solve_triangular(C, dRdetadomega, lower=True)
                    )
                )

                dreddetadomega_arg1 = (self.nt / sigma_2) * (
                    dsigma2detadomega - (1 / sigma_2) * dsigma[omega] * dsigma[eta]
                )
                dreddetadomega = (
                    -(dreddetadomega_arg1 - dreddetadomega_tr_1 + dreddetadomega_tr_2)
                    / self.nt
                )

                hess[ind_0 + i, 0] = (self.nt / np.log(10) * dreddetadomega).item()

                if self.name in ["MGP"] and eta == omega:
                    hess[ind_0 + i, 0] += log_prior[eta].item()
            par["Rinv_dR_gamma"] = Rinv_dRdomega_gamma_all
            par["Rinv_dmu"] = Rinv_dmudomega_all
        return hess, hess_ij, par

    def _predict_init(self, x, is_acting):
        if not (self.is_continuous):
            if is_acting is None:
                x, is_acting = self.design_space.correct_get_acting(x)
            n_eval, _ = x.shape
            _, ij = cross_distances(x, self.X_train)
            dx = gower_componentwise_distances(
                x,
                x_is_acting=is_acting,
                design_space=self.design_space,
                hierarchical_kernel=self.options["hierarchical_kernel"],
                y=np.copy(self.X_train),
                y_is_acting=self.is_acting_train,
            )
            listcatdecreed = self.design_space.is_conditionally_acting[
                self.cat_features
            ]
            if np.any(listcatdecreed):
                dx = self._correct_distances_cat_decreed(
                    dx,
                    is_acting,
                    listcatdecreed,
                    ij,
                    is_acting_y=self.is_acting_train,
                    mixint_type=MixIntKernelType.GOWER,
                )
            if self.options["categorical_kernel"] == MixIntKernelType.CONT_RELAX:
                Xpred, _ = self.design_space.unfold_x(x)
                Xpred_norma = (Xpred - self.X2_offset) / self.X2_scale
                dx = differences(Xpred_norma, Y=self.X2_norma.copy())
                listcatdecreed = self.design_space.is_conditionally_acting[
                    self.cat_features
                ]
                if np.any(listcatdecreed):
                    dx = self._correct_distances_cat_decreed(
                        dx,
                        is_acting,
                        listcatdecreed,
                        ij,
                        is_acting_y=self.is_acting_train,
                        mixint_type=MixIntKernelType.CONT_RELAX,
                    )

            Lij, _ = cross_levels(
                X=x, ij=ij, design_space=self.design_space, y=self.X_train
            )
            self.ij = ij
        else:
            n_eval, _ = x.shape
            X_cont = (np.copy(x) - self.X_offset) / self.X_scale
            dx = differences(X_cont, Y=self.X_norma.copy())
            ij = 0
            Lij = 0
        return x, is_acting, n_eval, ij, Lij, dx

    def predict_values(self, x: np.ndarray, is_acting=None) -> np.ndarray:
        """
        Predict the output values at a set of points.

        Parameters
        ----------
        x : np.ndarray[nt, nx] or np.ndarray[nt]
            Input values for the prediction points.
        is_acting : np.ndarray[nt, nx] or np.ndarray[nt]
            Matrix specifying for each design variable whether it is acting or not (for hierarchical design spaces)

        Returns
        -------
        y : np.ndarray[nt, ny]
            Output values at the prediction points.
        """
        x = ensure_2d_array(x, "x")
        self._check_xdim(x)

        if is_acting is not None:
            is_acting = ensure_2d_array(is_acting, "is_acting")
            if is_acting.shape != x.shape:
                raise ValueError(
                    f"is_acting should have the same dimensions as x: {is_acting.shape} != {x.shape}"
                )

        n = x.shape[0]
        x2 = np.copy(x)
        self.printer.active = (
            self.options["print_global"] and self.options["print_prediction"]
        )

        if self.name == "MixExp":
            # Mixture of experts model
            self.printer._title("Evaluation of the Mixture of experts")
        else:
            self.printer._title("Evaluation")
        self.printer("   %-12s : %i" % ("# eval points.", n))
        self.printer()

        # Evaluate the unknown points using the specified model-method
        with self.printer._timed_context("Predicting", key="prediction"):
            y = self._predict_values(x2, is_acting=is_acting)
        time_pt = self.printer._time("prediction")[-1] / n
        self.printer()
        self.printer("Prediction time/pt. (sec) : %10.7f" % time_pt)
        self.printer()
        return y.reshape((n, self.ny))

    def _predict_values(self, x: np.ndarray, is_acting=None) -> np.ndarray:
        """
        Evaluates the model at a set of points.

        Parameters
        ----------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        is_acting : np.ndarray[nt, nx] or np.ndarray[nt]
            Matrix specifying for each design variable whether it is acting or not (for hierarchical design spaces)

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        # Initialization
        if not (self.is_continuous):
            x, is_acting, n_eval, ij, Lij, dx = self._predict_init(x, is_acting)

            r = self._matrix_data_corr(
                corr=self.options["corr"],
                design_space=self.design_space,
                power=self.options["pow_exp_power"],
                theta=self.optimal_theta,
                theta_bounds=self.options["theta_bounds"],
                dx=dx,
                Lij=Lij,
                n_levels=self.n_levels,
                cat_features=self.cat_features,
                cat_kernel=self.options["categorical_kernel"],
                x=x,
            ).reshape(n_eval, self.nt)

            X_cont, _ = compute_X_cont(x, self.design_space)

        else:
            _, _, n_eval, _, _, dx = self._predict_init(x, is_acting)
            X_cont = np.copy(x)
            d = self._componentwise_distance(dx)
            # Compute the correlation function
            r = self._correlation_types[self.options["corr"]](
                self.optimal_theta, d
            ).reshape(n_eval, self.nt)
            y = np.zeros(n_eval)
        X_cont = (X_cont - self.X_offset) / self.X_scale
        # Compute the regression function
        f = self._regression_types[self.options["poly"]](X_cont)
        # Scaled predictor
        y_ = np.dot(f, self.optimal_par["beta"]) + np.dot(r, self.optimal_par["gamma"])
        # Predictor
        y = (self.y_mean + self.y_std * y_).ravel()
        return y

    def _predict_derivatives(self, x, kx):
        """
        Evaluates the derivatives at a set of points.

        Parameters
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        kx : int
            The 0-based index of the input variable with respect to which derivatives are desired.

        Returns
        -------
        y : np.ndarray
            Derivative values.
        """
        # Initialization
        n_eval, _ = x.shape

        x = (x - self.X_offset) / self.X_scale
        # Get pairwise componentwise L1-distances to the input training set

        dx = differences(x, Y=self.X_norma.copy())
        d = self._componentwise_distance(dx)

        dd = self._componentwise_distance(
            dx, theta=self.optimal_theta, return_derivative=True
        )

        # Compute the correlation function
        derivative_dic = {"dx": dx, "dd": dd}

        r, dr = self._correlation_types[self.options["corr"]](
            self.optimal_theta, d, derivative_params=derivative_dic
        )
        r = r.reshape(n_eval, self.nt)

        drx = dr[:, kx].reshape(n_eval, self.nt)

        if self.options["poly"] == "constant":
            df = np.zeros((1, self.nx))
        elif self.options["poly"] == "linear":
            df = np.zeros((self.nx + 1, self.nx))
            df[1:, :] = np.eye(self.nx)
        else:
            raise ValueError(
                "The derivative is only available for ordinary kriging or "
                + "universal kriging using a linear trend"
            )
        # Beta and gamma = R^-1(y-FBeta)
        beta = self.optimal_par["beta"]
        gamma = self.optimal_par["gamma"]
        df_dx = np.dot(df.T, beta)

        y = (df_dx[kx] + np.dot(drx, gamma)) * self.y_std / self.X_scale[kx]
        return y

    def predict_variances(self, x: np.ndarray, is_acting=None) -> np.ndarray:
        """
        Predict the variances at a set of points.

        Parameters
        ----------
        x : np.ndarray[nt, nx] or np.ndarray[nt]
            Input values for the prediction points.
        is_acting : np.ndarray[nt, nx] or np.ndarray[nt]
            Matrix specifying for each design variable whether it is acting or not (for hierarchical design spaces)

        Returns
        -------
        s2 : np.ndarray[nt, ny]
            Variances.
        """
        check_support(self, "variances")
        x = ensure_2d_array(x, "x")
        self._check_xdim(x)

        if is_acting is not None:
            is_acting = ensure_2d_array(is_acting, "is_acting")
            if is_acting.shape != x.shape:
                raise ValueError(
                    f"is_acting should have the same dimensions as x: {is_acting.shape} != {x.shape}"
                )

        n = x.shape[0]
        x2 = np.copy(x)
        s2 = self._predict_variances(x2, is_acting=is_acting)
        return s2.reshape((n, self.ny))

    def _predict_variances(self, x: np.ndarray, is_acting=None) -> np.ndarray:
        """
        Provide uncertainty of the model at a set of points
        Parameters
        ----------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        is_acting : np.ndarray[nt, nx] or np.ndarray[nt]
            Matrix specifying for each design variable whether it is acting or not (for hierarchical design spaces)
        Returns
        -------
        MSE : np.ndarray
            Evaluation point output variable MSE
        """
        # Initialization
        if not (self.is_continuous):
            x, is_acting, n_eval, ij, Lij, dx = self._predict_init(x, is_acting)
            X_cont = x

            r = self._matrix_data_corr(
                corr=self.options["corr"],
                design_space=self.design_space,
                power=self.options["pow_exp_power"],
                theta=self.optimal_theta,
                theta_bounds=self.options["theta_bounds"],
                dx=dx,
                Lij=Lij,
                n_levels=self.n_levels,
                cat_features=self.cat_features,
                cat_kernel=self.options["categorical_kernel"],
                x=x,
            ).reshape(n_eval, self.nt)

            X_cont, _ = compute_X_cont(x, self.design_space)
        else:
            _, _, n_eval, _, _, dx = self._predict_init(x, is_acting)
            X_cont = np.copy(x)
            d = self._componentwise_distance(dx)
            # Compute the correlation function
            r = self._correlation_types[self.options["corr"]](
                self.optimal_theta, d
            ).reshape(n_eval, self.nt)
        X_cont = (X_cont - self.X_offset) / self.X_scale
        C = self.optimal_par["C"]
        rt = linalg.solve_triangular(C, r.T, lower=True)

        u = linalg.solve_triangular(
            self.optimal_par["G"].T,
            np.dot(self.optimal_par["Ft"].T, rt)
            - self._regression_types[self.options["poly"]](X_cont).T,
        )
        A = self.optimal_par["sigma2"]
        B = 1.0 - (rt**2.0).sum(axis=0) + (u**2.0).sum(axis=0)
        # machine precision: force to zero!
        B[B < 1e-12] = 0
        MSE = np.einsum("i,j -> ji", A, B)
        # Mean Squared Error might be slightly negative depending on
        # machine precision: force to zero!
        MSE[MSE < 0.0] = 0.0
        return MSE

    def _predict_variance_derivatives(self, x, kx):
        """
        Provide the derivative of the variance of the model at a set of points
        Parameters
        -----------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        kx : int
            The 0-based index of the input variable with respect to which derivatives are desired.

        Returns
        -------
         derived_variance:  np.ndarray
             The jacobian of the variance of the kriging model (the kx-th derivative)
        """

        # Initialization
        n_eval, _ = x.shape
        x = (x - self.X_offset) / self.X_scale
        theta = self.optimal_theta
        # Get pairwise componentwise L1-distances to the input training set
        dx = differences(x, Y=self.X_norma.copy())
        d = self._componentwise_distance(dx)
        dd = self._componentwise_distance(
            dx, theta=self.optimal_theta, return_derivative=True
        )
        derivative_dic = {"dx": dx, "dd": dd}

        sigma2 = self.optimal_par["sigma2"]
        C = self.optimal_par["C"]

        # p1 : derivative of (rt**2.0).sum(axis=0)
        r, dr = self._correlation_types[self.options["corr"]](
            theta, d, derivative_params=derivative_dic
        )
        r = r.reshape(n_eval, self.nt)
        drx = dr[:, kx]
        drx = drx.reshape(n_eval, self.nt)

        rt = linalg.solve_triangular(C, r.T, lower=True)
        invKr = linalg.solve_triangular(C.T, rt)
        p1 = 2 * np.dot(drx, invKr).T

        # p2 : derivative of (u**2.0).sum(axis=0)
        f_x = self._regression_types[self.options["poly"]](x).T
        F = self.F
        rho2 = linalg.solve_triangular(C, F, lower=True)
        invKF = linalg.solve_triangular(C.T, rho2)

        A = f_x.T - np.dot(r, invKF)

        B = np.dot(F.T, invKF)
        rho3 = linalg.cholesky(B, lower=True)
        invBAt = linalg.solve_triangular(rho3, A.T, lower=True)
        D = linalg.solve_triangular(rho3.T, invBAt)

        if self.options["poly"] == "constant":
            df = np.zeros((1, self.nx))
        elif self.options["poly"] == "linear":
            df = np.zeros((self.nx + 1, self.nx))
            df[1:, :] = np.eye(self.nx)
        else:
            raise ValueError(
                "The derivative is only available for ordinary kriging or "
                + "universal kriging using a linear trend"
            )

        dA = df[:, kx].T - np.dot(drx, invKF)
        p3 = 2 * np.dot(dA, D).T

        # prime : derivative of MSE
        # MSE ~1.0 - (rt**2.0).sum(axis=0) + (u**2.0).sum(axis=0)
        prime = 0 - p1 + p3
        ## scaling factors
        x_std = self.X_scale[kx]
        derived_variance = np.array((np.outer(sigma2, np.diag(prime.T)) / x_std))

        return np.atleast_2d(derived_variance.T)

    def _optimize_hyperparam(self, D):
        """
        This function evaluates the Gaussian Process model at x.

        Parameters
        ----------
        D: np.ndarray [n_obs * (n_obs - 1) / 2, dim]
            - The componentwise cross-spatial-correlation-distance between the
              vectors in X.
           For SGP surrogate, D is not used

        Returns
        -------
        best_optimal_rlf_value: real
            - The value of the reduced likelihood function associated to the
              best autocorrelation parameters theta.
        best_optimal_par: dict()
            - A dictionary containing the requested Gaussian Process model
              parameters.
        best_optimal_theta: list(n_comp) or list(dim)
            - The best hyperparameters found by the optimization.
        """
        # reinitialize optimization best values
        self.best_iteration_fail = None
        self._thetaMemory = None
        # Initialize the hyperparameter-optimization
        if self.name in ["MGP"]:

            def minus_reduced_likelihood_function(theta):
                res = -self._reduced_likelihood_function(theta)[0]
                return res

            def grad_minus_reduced_likelihood_function(theta):
                grad = -self._reduced_likelihood_gradient(theta)[0]
                return grad

            def hessian_minus_reduced_likelihood_function(theta):
                hess = -self._reduced_likelihood_hessian(theta)[0]
                return hess

        else:

            def minus_reduced_likelihood_function(log10t):
                return -self._reduced_likelihood_function(theta=10.0**log10t)[0]

            def grad_minus_reduced_likelihood_function(log10t):
                log10t_2d = np.atleast_2d(log10t).T
                res = (
                    -np.log(10.0)
                    * (10.0**log10t_2d)
                    * (self._reduced_likelihood_gradient(10.0**log10t_2d)[0])
                )
                return res

            def hessian_minus_reduced_likelihood_function(log10t):
                log10t_2d = np.atleast_2d(log10t).T
                res = (
                    -np.log(10.0)
                    * (10.0**log10t_2d)
                    * (self._reduced_likelihood_hessian(10.0**log10t_2d)[0])
                )
                return res

        limit, _rhobeg = max(12 * len(self.options["theta0"]), 50), 0.5
        exit_function = False
        if "KPLSK" in self.name:
            n_iter = 1
        else:
            n_iter = 0

        (
            best_optimal_theta,
            best_optimal_rlf_value,
            best_optimal_par,
            constraints,
        ) = (
            [],
            [],
            [],
            [],
        )

        for ii in range(n_iter, -1, -1):
            bounds_hyp = []

            self.theta0 = deepcopy(self.options["theta0"])
            for i in range(len(self.theta0)):
                # In practice, in 1D and for X in [0,1], theta^{-2} in [1e-2,infty),
                # i.e. theta in (0,1e1], is a good choice to avoid overfitting.
                # By standardising X in R, X_norm = (X-X_mean)/X_std, then
                # X_norm in [-1,1] if considering one std intervals. This leads
                # to theta in (0,2e1]
                theta_bounds = self.options["theta_bounds"]
                if self.theta0[i] < theta_bounds[0] or self.theta0[i] > theta_bounds[1]:
                    if ii == 0 and "KPLSK" in self.name:
                        if self.theta0[i] - theta_bounds[1] > 0:
                            self.theta0[i] = theta_bounds[1] - 1e-10
                        else:
                            self.theta0[i] = theta_bounds[0] + 1e-10
                    else:
                        warnings.warn(
                            f"theta0 is out the feasible bounds ({self.theta0}[{i}] out of \
                                [{theta_bounds[0]}, {theta_bounds[1]}]). \
                                    A random initialisation is used instead."
                        )
                        self.theta0[i] = self.random_state.rand()
                        self.theta0[i] = (
                            self.theta0[i] * (theta_bounds[1] - theta_bounds[0])
                            + theta_bounds[0]
                        )

                if self.name in ["MGP"]:  # to be discussed with R. Priem
                    constraints.append(lambda theta, i=i: theta[i] + theta_bounds[1])
                    constraints.append(lambda theta, i=i: theta_bounds[1] - theta[i])
                    bounds_hyp.append((-theta_bounds[1], theta_bounds[1]))
                else:
                    log10t_bounds = np.log10(theta_bounds)
                    constraints.append(lambda log10t, i=i: log10t[i] - log10t_bounds[0])
                    constraints.append(lambda log10t, i=i: log10t_bounds[1] - log10t[i])
                    bounds_hyp.append(log10t_bounds)

            if self.name in ["MGP"]:
                theta0_rand = m_norm.rvs(
                    self.options["prior"]["mean"] * len(self.theta0),
                    self.options["prior"]["var"],
                    1,
                )
                theta0 = self.theta0
            else:
                theta_bounds = self.options["theta_bounds"]
                log10t_bounds = np.log10(theta_bounds)
                theta0_rand = self.random_state.rand(len(self.theta0))
                theta0_rand = (
                    theta0_rand * (log10t_bounds[1] - log10t_bounds[0])
                    + log10t_bounds[0]
                )
                theta0 = np.log10(self.theta0)

            if self.name not in ["SGP"]:
                if not (self.is_continuous):
                    self.D = D
                else:
                    ##from abs distance to kernel distance
                    self.D = self._componentwise_distance(D, opt=ii)
            else:  # SGP case, D is not used
                pass

            # Initialization
            k, incr, stop, best_optimal_rlf_value, max_retry = 0, 0, 1, -1e20, 10
            while k < stop:
                # Use specified starting point as first guess
                self.noise0 = np.array(self.options["noise0"])
                noise_bounds = self.options["noise_bounds"]

                # SGP: GP variance is optimized too
                offset = 0
                if self.name in ["SGP"]:
                    sigma2_0 = np.log10(np.array([self.y_std[0] ** 2]))
                    theta0_sigma2 = np.concatenate([theta0, sigma2_0])
                    sigma2_bounds = np.log10(
                        np.array([1e-12, (3.0 * self.y_std[0]) ** 2])
                    )
                    constraints.append(
                        lambda log10t: log10t[len(self.theta0)] - sigma2_bounds[0]
                    )
                    constraints.append(
                        lambda log10t: sigma2_bounds[1] - log10t[len(self.theta0)]
                    )
                    bounds_hyp.append(sigma2_bounds)
                    offset = 1
                    theta0 = theta0_sigma2
                    theta0_rand = np.concatenate([theta0_rand, sigma2_0])

                if self.options["eval_noise"] and not self.options["use_het_noise"]:
                    self.noise0[self.noise0 == 0.0] = noise_bounds[0]
                    for i in range(len(self.noise0)):
                        if (
                            self.noise0[i] < noise_bounds[0]
                            or self.noise0[i] > noise_bounds[1]
                        ):
                            self.noise0[i] = noise_bounds[0]
                            warnings.warn(
                                "Warning: noise0 is out the feasible bounds. The lowest possible value is used instead."
                            )

                    theta0 = np.concatenate(
                        [theta0, np.log10(np.array([self.noise0]).flatten())]
                    )
                    theta0_rand = np.concatenate(
                        [
                            theta0_rand,
                            np.log10(np.array([self.noise0]).flatten()),
                        ]
                    )

                    for i in range(len(self.noise0)):
                        noise_bounds = np.log10(noise_bounds)
                        constraints.append(
                            lambda log10t, i=i: log10t[offset + i + len(self.theta0)]
                            - noise_bounds[0]
                        )
                        constraints.append(
                            lambda log10t, i=i: noise_bounds[1]
                            - log10t[offset + i + len(self.theta0)]
                        )
                        bounds_hyp.append(noise_bounds)
                theta_limits = np.repeat(
                    np.log10([theta_bounds]), repeats=len(theta0), axis=0
                )
                theta_all_loops = np.vstack((theta0, theta0_rand))
                if ii == 1 or "KPLSK" not in self.name:
                    if self.options["n_start"] > 1:
                        sampling = LHS(
                            xlimits=theta_limits,
                            criterion="maximin",
                            random_state=self.random_state,
                        )
                        theta_lhs_loops = sampling(self.options["n_start"])
                        theta_all_loops = np.vstack((theta_all_loops, theta_lhs_loops))

                optimal_theta_res = {"fun": float("inf")}
                optimal_theta_res_loop = None
                try:
                    if self.options["hyper_opt"] == "Cobyla":
                        for theta0_loop in theta_all_loops:
                            optimal_theta_res_loop = optimize.minimize(
                                minus_reduced_likelihood_function,
                                theta0_loop,
                                constraints=[
                                    {"fun": con, "type": "ineq"} for con in constraints
                                ],
                                method="COBYLA",
                                options={
                                    "rhobeg": _rhobeg,
                                    "tol": 1e-4,
                                    "maxiter": limit,
                                },
                            )
                            if optimal_theta_res_loop["fun"] < optimal_theta_res["fun"]:
                                optimal_theta_res = optimal_theta_res_loop

                    elif self.options["hyper_opt"] == "TNC":
                        if self.options["use_het_noise"]:
                            raise ValueError(
                                "For heteroscedastic noise, please use Cobyla"
                            )
                        for theta0_loop in theta_all_loops:
                            optimal_theta_res_loop = optimize.minimize(
                                minus_reduced_likelihood_function,
                                theta0_loop,
                                method="TNC",
                                jac=grad_minus_reduced_likelihood_function,
                                ###The hessian information is available but never used
                                #
                                ####hess=hessian_minus_reduced_likelihood_function,
                                bounds=bounds_hyp,
                                options={"maxfun": limit},
                            )
                            if optimal_theta_res_loop["fun"] < optimal_theta_res["fun"]:
                                optimal_theta_res = optimal_theta_res_loop

                    if "x" not in optimal_theta_res:
                        raise ValueError(
                            f"Optimizer encountered a problem: {optimal_theta_res_loop!s}"
                        )
                    optimal_theta = optimal_theta_res["x"]

                    if self.name not in ["MGP"]:
                        optimal_theta = 10**optimal_theta

                    optimal_rlf_value, optimal_par = self._reduced_likelihood_function(
                        theta=optimal_theta
                    )
                    # Compare the new optimizer to the best previous one
                    if k > 0:
                        if np.isinf(optimal_rlf_value):
                            stop += 1
                            if incr != 0:
                                return
                            if stop > max_retry:
                                raise ValueError(
                                    "%d attempts to train the model failed" % max_retry
                                )
                        else:
                            if optimal_rlf_value >= self.best_iteration_fail:
                                if optimal_rlf_value > best_optimal_rlf_value:
                                    best_optimal_rlf_value = optimal_rlf_value
                                    best_optimal_par = optimal_par
                                    best_optimal_theta = optimal_theta
                                else:
                                    if (
                                        self.best_iteration_fail
                                        > best_optimal_rlf_value
                                    ):
                                        best_optimal_theta = self._thetaMemory
                                        (
                                            best_optimal_rlf_value,
                                            best_optimal_par,
                                        ) = self._reduced_likelihood_function(
                                            theta=best_optimal_theta
                                        )
                    else:
                        if np.isinf(optimal_rlf_value):
                            stop += 1
                        else:
                            best_optimal_rlf_value = optimal_rlf_value
                            best_optimal_par = optimal_par
                            best_optimal_theta = optimal_theta
                    k += 1
                except ValueError as ve:
                    # raise ve
                    # If iteration is max when fmin_cobyla fail is not reached
                    if self.nb_ill_matrix > 0:
                        self.nb_ill_matrix -= 1
                        k += 1
                        stop += 1
                        # One evaluation objectif function is done at least
                        if self.best_iteration_fail is not None:
                            if self.best_iteration_fail > best_optimal_rlf_value:
                                best_optimal_theta = self._thetaMemory
                                (
                                    best_optimal_rlf_value,
                                    best_optimal_par,
                                ) = self._reduced_likelihood_function(
                                    theta=best_optimal_theta
                                )
                    # Optimization fail
                    elif np.size(best_optimal_par) == 0:
                        print("Optimization failed. Try increasing the ``nugget``")
                        raise ve
                    # Break the while loop
                    else:
                        k = stop + 1
                        print("fmin_cobyla failed but the best value is retained")

            if "KPLSK" in self.name:
                if self.options["eval_noise"]:
                    # best_optimal_theta contains [theta, noise] if eval_noise = True
                    theta = best_optimal_theta[:-1]
                else:
                    # best_optimal_theta contains [theta] if eval_noise = False
                    theta = best_optimal_theta

                if exit_function:
                    return best_optimal_rlf_value, best_optimal_par, best_optimal_theta

                if self.options["corr"] == "squar_exp":
                    self.options["theta0"] = (theta * self.coeff_pls**2).sum(1)
                else:
                    self.options["theta0"] = (theta * np.abs(self.coeff_pls)).sum(1)

                self.options["n_comp"] = int(self.nx)
                limit = 10 * self.options["n_comp"]
                self.best_iteration_fail = None
                exit_function = True
        return best_optimal_rlf_value, best_optimal_par, best_optimal_theta

    def _check_param(self):
        """
        This function checks some parameters of the model
        and amend theta0 if possible (see _amend_theta0_option).
        """
        d = self.options["n_comp"] if "n_comp" in self.options else self.nx
        if self.name in ["KPLS"]:
            if self.options["corr"] not in ["pow_exp", "squar_exp", "abs_exp"]:
                raise ValueError(
                    "KPLS only works with a squared exponential, or an absolute exponential kernel with variable power"
                )
            if (
                self.options["categorical_kernel"]
                not in [
                    MixIntKernelType.EXP_HOMO_HSPHERE,
                    MixIntKernelType.HOMO_HSPHERE,
                ]
                and self.name == "KPLS"
            ):
                if self.options["cat_kernel_comps"] is not None:
                    raise ValueError(
                        "cat_kernel_comps option is for homoscedastic kernel."
                    )

        mat_dim = (
            self.options["cat_kernel_comps"]
            if "cat_kernel_comps" in self.options
            else None
        )

        n_comp = self.options["n_comp"] if "n_comp" in self.options else None
        n_param = compute_n_param(
            self.design_space,
            self.options["categorical_kernel"],
            d,
            n_comp,
            mat_dim,
        )

        if self.options["corr"] == "squar_sin_exp":
            if (
                self.is_continuous
                or self.options["categorical_kernel"] == MixIntKernelType.GOWER
            ):
                self.options["theta0"] *= np.ones(2 * n_param)
            else:
                n_param += len([self.design_space.is_cat_mask])
                self.options["theta0"] *= np.ones(n_param)

        else:
            self.options["theta0"] *= np.ones(n_param)
        if (
            self.options["corr"] not in ["squar_exp", "abs_exp", "pow_exp"]
            and not (self.is_continuous)
            and self.options["categorical_kernel"]
            not in [
                MixIntKernelType.GOWER,
                MixIntKernelType.COMPOUND_SYMMETRY,
                MixIntKernelType.HOMO_HSPHERE,
            ]
        ):
            raise ValueError(
                "Categorical kernels should be matrix or exponential based."
            )

        if len(self.options["theta0"]) != d and (
            self.options["categorical_kernel"]
            in [MixIntKernelType.GOWER, MixIntKernelType.COMPOUND_SYMMETRY]
            or self.is_continuous
        ):
            if len(self.options["theta0"]) == 1:
                self.options["theta0"] *= np.ones(d)
            else:
                if self.options["corr"] != "squar_sin_exp":
                    raise ValueError(
                        "the length of theta0 (%s) should be equal to the number of dim (%s)."
                        % (len(self.options["theta0"]), d)
                    )
        if (
            self.options["eval_noise"] or np.max(self.options["noise0"]) > 1e-12
        ) and self.options["hyper_opt"] == "TNC":
            warnings.warn(
                "TNC not available yet for noise handling. Switching to Cobyla"
            )
            self.options["hyper_opt"] = "Cobyla"

        if self.options["use_het_noise"] and not self.options["eval_noise"]:
            if len(self.options["noise0"]) != self.nt:
                if len(self.options["noise0"]) == 1:
                    self.options["noise0"] *= np.ones(self.nt)
                else:
                    raise ValueError(
                        "for the heteroscedastic case, the length of noise0 (%s) \
                            should be equal to the number of observations (%s)."
                        % (len(self.options["noise0"]), self.nt)
                    )
        if not self.options["use_het_noise"]:
            if len(self.options["noise0"]) != 1:
                raise ValueError(
                    "for the homoscedastic noise case, the length of noise0 (%s) should be equal to one."
                    % (len(self.options["noise0"]))
                )

        if self.supports["training_derivatives"]:
            if 1 not in self.training_points[None]:
                raise Exception("Derivative values are needed for using the GEK model.")

    def _check_F(self, n_samples_F, p):
        """
        This function check the F-parameters of the model.
        """

        if n_samples_F != self.nt:
            raise Exception(
                "Number of rows in F and X do not match. Most "
                "likely something is going wrong with the "
                "regression model."
            )
        if p > n_samples_F:
            raise Exception(
                (
                    "Ordinary least squares problem is undetermined "
                    "n_samples=%d must be greater than the "
                    "regression model size p=%d."
                )
                % (self.nt, p)
            )


def compute_n_param(design_space, cat_kernel, d, n_comp, mat_dim):
    """
    Returns the he number of parameters needed for an homoscedastic or full group kernel.
    Parameters
     ----------
    design_space: BaseDesignSpace
            - design space definition
    cat_kernel : string
            -The kernel to use for categorical inputs. Only for non continuous Kriging,
    d: int
            - n_comp or nx
    n_comp : int
            - if PLS, then it is the number of components else None,
    mat_dim : int
            - if PLS, then it is the number of components for matrix kernel (mixed integer) else None,
    Returns
    -------
     n_param: int
            - The number of parameters.
    """
    n_param = design_space.n_dv
    if n_comp is not None:
        n_param = d
        if cat_kernel == MixIntKernelType.CONT_RELAX:
            return n_param
        if mat_dim is not None:
            return int(np.sum([i * (i - 1) / 2 for i in mat_dim]) + n_param)
    if cat_kernel in [MixIntKernelType.GOWER, MixIntKernelType.COMPOUND_SYMMETRY]:
        return n_param
    for i, dv in enumerate(design_space.design_variables):
        if isinstance(dv, CategoricalVariable):
            n_values = dv.n_values
            if design_space.n_dv == d:
                n_param -= 1
            if cat_kernel in [
                MixIntKernelType.EXP_HOMO_HSPHERE,
                MixIntKernelType.HOMO_HSPHERE,
            ]:
                n_param += int(n_values * (n_values - 1) / 2)
            if cat_kernel == MixIntKernelType.CONT_RELAX:
                n_param += int(n_values)
    return n_param



"""
Author: P.Saves and J.H. Bussemaker
This package is distributed under New BSD license.

Cantilever beam problem from:
P. Saves, Y. Diouane, N. Bartoli, T. Lefebvre, and J. Morlier. A mixed-categorical correlation kernel
for gaussian process, 2022
"""

import numpy as np

from smt.problems.problem import Problem
from smt.utils.design_space import (
    CategoricalVariable,
    DesignSpace,
    FloatVariable,
    IntegerVariable,
    OrdinalVariable,
)


class HierarchicalGoldstein(Problem):
    def _setup(self):
        ds = DesignSpace(
            [
                CategoricalVariable(values=[0, 1, 2, 3]),  # meta
                OrdinalVariable(values=[0, 1]),  # x1
                FloatVariable(0, 100),  # x2
                FloatVariable(0, 100),
                FloatVariable(0, 100),
                FloatVariable(0, 100),
                FloatVariable(0, 100),
                IntegerVariable(0, 2),  # x7
                IntegerVariable(0, 2),
                IntegerVariable(0, 2),
                IntegerVariable(0, 2),
            ]
        )

        # x4 is acting if meta == 1, 3
        ds.declare_decreed_var(decreed_var=4, meta_var=0, meta_value=[1, 3])
        # x5 is acting if meta == 2, 3
        ds.declare_decreed_var(decreed_var=5, meta_var=0, meta_value=[2, 3])
        # x7 is acting if meta == 0, 2
        ds.declare_decreed_var(decreed_var=7, meta_var=0, meta_value=[0, 2])
        # x8 is acting if meta == 0, 1
        ds.declare_decreed_var(decreed_var=8, meta_var=0, meta_value=[0, 1])

        self._set_design_space(ds)

    def _evaluate(self, x: np.ndarray, kx=0) -> np.ndarray:
        def H(x1, x2, x3, x4, z3, z4, x5, cos_term):
            h = (
                53.3108
                + 0.184901 * x1
                - 5.02914 * x1**3 * 10 ** (-6)
                + 7.72522 * x1**z3 * 10 ** (-8)
                - 0.0870775 * x2
                - 0.106959 * x3
                + 7.98772 * x3**z4 * 10 ** (-6)
                + 0.00242482 * x4
                + 1.32851 * x4**3 * 10 ** (-6)
                - 0.00146393 * x1 * x2
                - 0.00301588 * x1 * x3
                - 0.00272291 * x1 * x4
                + 0.0017004 * x2 * x3
                + 0.0038428 * x2 * x4
                - 0.000198969 * x3 * x4
                + 1.86025 * x1 * x2 * x3 * 10 ** (-5)
                - 1.88719 * x1 * x2 * x4 * 10 ** (-6)
                + 2.50923 * x1 * x3 * x4 * 10 ** (-5)
                - 5.62199 * x2 * x3 * x4 * 10 ** (-5)
            )
            if cos_term:
                h += 5.0 * np.cos(2.0 * np.pi * (x5 / 100.0)) - 2.0
            return h

        def f1(x1, x2, z1, z2, z3, z4, x5, cos_term):
            c1 = z1 == 0
            c2 = z1 == 1
            c3 = z1 == 2

            c4 = z2 == 0
            c5 = z2 == 1
            c6 = z2 == 2

            y = (
                c4
                * (
                    c1 * H(x1, x2, 20, 20, z3, z4, x5, cos_term)
                    + c2 * H(x1, x2, 50, 20, z3, z4, x5, cos_term)
                    + c3 * H(x1, x2, 80, 20, z3, z4, x5, cos_term)
                )
                + c5
                * (
                    c1 * H(x1, x2, 20, 50, z3, z4, x5, cos_term)
                    + c2 * H(x1, x2, 50, 50, z3, z4, x5, cos_term)
                    + c3 * H(x1, x2, 80, 50, z3, z4, x5, cos_term)
                )
                + c6
                * (
                    c1 * H(x1, x2, 20, 80, z3, z4, x5, cos_term)
                    + c2 * H(x1, x2, 50, 80, z3, z4, x5, cos_term)
                    + c3 * H(x1, x2, 80, 80, z3, z4, x5, cos_term)
                )
            )
            return y

        def f2(x1, x2, x3, z2, z3, z4, x5, cos_term):
            c4 = z2 == 0
            c5 = z2 == 1
            c6 = z2 == 2

            y = (
                c4 * H(x1, x2, x3, 20, z3, z4, x5, cos_term)
                + c5 * H(x1, x2, x3, 50, z3, z4, x5, cos_term)
                + c6 * H(x1, x2, x3, 80, z3, z4, x5, cos_term)
            )
            return y

        def f3(x1, x2, x4, z1, z3, z4, x5, cos_term):
            c1 = z1 == 0
            c2 = z1 == 1
            c3 = z1 == 2

            y = (
                c1 * H(x1, x2, 20, x4, z3, z4, x5, cos_term)
                + c2 * H(x1, x2, 50, x4, z3, z4, x5, cos_term)
                + c3 * H(x1, x2, 80, x4, z3, z4, x5, cos_term)
            )
            return y

        y = []
        for xi in x:
            if xi[0] == 0:
                y.append(
                    f1(xi[2], xi[3], xi[7], xi[8], xi[9], xi[10], xi[6], cos_term=xi[1])
                )
            elif xi[0] == 1:
                y.append(
                    f2(xi[2], xi[3], xi[4], xi[8], xi[9], xi[10], xi[6], cos_term=xi[1])
                )
            elif xi[0] == 2:
                y.append(
                    f3(xi[2], xi[3], xi[5], xi[7], xi[9], xi[10], xi[6], cos_term=xi[1])
                )
            elif xi[0] == 3:
                y.append(
                    H(xi[2], xi[3], xi[4], xi[5], xi[9], xi[10], xi[6], cos_term=xi[1])
                )
            else:
                raise ValueError(f"Unexpected x0: {xi[0]}")
        return np.array(y)



from .ls import LS
from .qp import QP
from .krg import KRG
from .kpls import KPLS
from .gekpls import GEKPLS
from .kplsk import KPLSK
from .gpx import GPX
from .genn import GENN
from .mgp import MGP
from .sgp import SGP

from .krg_based import MixIntKernelType
from smt.utils.design_space import (
    DesignSpace,
    FloatVariable,
    IntegerVariable,
    OrdinalVariable,
    CategoricalVariable,
)
from smt.utils.kriging import MixHrcKernelType

__all__ = [
    "LS",
    "QP",
    "KRG",
    "KPLS",
    "GEKPLS",
    "KPLSK",
    "GPX",
    "GENN",
    "MGP",
    "SGP",
    "MixIntKernelType",
    "DesignSpace",
    "FloatVariable",
    "IntegerVariable",
    "OrdinalVariable",
    "CategoricalVariable",
    "MixHrcKernelType",
]

try:
    from .idw import IDW
    from .rbf import RBF
    from .rmtc import RMTC
    from .rmtb import RMTB

    __all__ = __all__ + ["IDW", "RBF", "RMTC", "RMTB"]

except ImportError:
    pass



"""
Gradient-Enhanced Neural Networks (GENN)

Author: Steven H. Berguin <steven.berguin@gtri.gatech.edu>

This package is distributed under New BSD license.
"""

from smt.surrogate_models.surrogate_model import SurrogateModel

import numpy as np
from jenn.model import NeuralNet
from typing import Union, Dict, Tuple, List


# The missing type
SMTrainingPoints = Dict[Union[int, None], Dict[int, List[np.ndarray]]]


def _smt_to_genn(
    training_points: SMTrainingPoints,
) -> Tuple[np.ndarray, np.ndarray, Union[np.ndarray, None]]:
    """Translate data structure from SMT to GENN.

    :param training_points: training data (as per SMT API)
    :return: X, array of shape (n_x, m) where n_x = number of inputs, m = number of examples
    :return: Y, array of shape (n_y, m) where n_y = number of outputs
    :return: J, array of shape (n_y, n_x, m)
    """
    # Retrieve training data from SMT training_points
    xt, yt = training_points[None][
        0
    ]  # training_points[name][0] = [np.array(xt), np.array(yt)]

    # Deduce number of dimensions and training examples
    m, n_x = xt.shape
    _, n_y = yt.shape

    # Assign training data but transpose to match neural net implementation
    X = xt
    Y = yt

    if len(training_points[None]) == 1:
        return X.T, Y.T, None

    # Loop to retrieve each partial derivative from SMT training_points
    J = np.zeros((m, n_x, n_y))
    for k in range(0, n_x):
        xt, dyt_dxt = training_points[None][k + 1]

        # assert that dimensions match
        assert xt.shape[0] == m
        assert xt.shape[1] == n_x
        assert dyt_dxt.shape[0] == m
        assert dyt_dxt.shape[1] == n_y

        # Assert that derivatives provided are for the same training points
        assert xt.all() == X.all()

        # Assign training derivative but transpose to match neural net implementation
        J[:, k, :] = dyt_dxt

    return X.T, Y.T, J.T


class GENN(SurrogateModel):
    """Gradient-Enhanced Neural Net."""

    def load_data(self, xt, yt, dyt_dxt=None):
        """Load all training data into surrogate model in one step.

        :param model: SurrogateModel object for which to load training data
        :param xt: smt data points at which response is evaluated
        :param yt: response at xt
        :param dyt_dxt: gradient at xt
        """
        m, n_x = (xt.size, 1) if xt.ndim <= 1 else xt.shape
        m, n_y = (yt.size, 1) if yt.ndim <= 1 else yt.shape

        # Reshape arrays
        xt = xt.reshape((m, n_x))
        yt = yt.reshape((m, n_y))

        # Load values
        self.set_training_values(xt, yt)

        # Load partials
        if dyt_dxt is not None:
            dyt_dxt = dyt_dxt.reshape((m, n_x))
            for i in range(n_x):
                self.set_training_derivatives(xt, dyt_dxt[:, i].reshape((m, 1)), i)

    @property
    def name(self) -> str:
        return "GENN"

    def _initialize(self):
        self.supports["derivatives"] = True
        self.supports["training_derivatives"] = True

        self.options.declare(
            "alpha",
            default=0.05,
            types=(int, float),
            desc="optimizer learning rate",
        )
        self.options.declare(
            "beta1",
            default=0.9,
            types=(int, float),
            desc="Adam optimizer tuning parameter",
        )
        self.options.declare(
            "beta2",
            default=0.99,
            types=(int, float),
            desc="Adam optimizer tuning parameter",
        )
        self.options.declare(
            "lambd",
            default=0.01,
            types=(int, float),
            desc="regularization coefficient",
        )
        self.options.declare(
            "gamma",
            default=1.0,
            types=(int, float),
            desc="gradient-enhancement coefficient",
        )
        self.options.declare(
            "hidden_layer_sizes",
            default=[12, 12],
            types=list,
            desc="number of nodes per hidden layer",
        )
        self.options.declare(
            "mini_batch_size",
            default=-1,
            types=int,
            desc="split data into batches of specified size",
        )
        self.options.declare(
            "num_epochs",
            default=1,
            types=int,
            desc="number of random passes through the data",
        )
        self.options.declare(
            "num_iterations",
            default=1000,
            types=int,
            desc="number of optimizer iterations per mini-batch",
        )
        self.options.declare(
            "seed",
            default=-1,
            types=int,
            desc="random seed to control repeatability",
        )
        self.options.declare(
            "is_print",
            default=False,
            types=bool,
            desc="print progress (or not)",
        )
        self.options.declare(
            "is_normalize",
            default=False,
            types=bool,
            desc="normalize training by mean and variance",
        )
        self.options.declare(
            "is_backtracking",
            default=False,
            types=bool,
            desc="refine step step during line search (fixed otherwise)",
        )

    def _final_initialize(self):
        inputs = [1]  # will be overwritten during training (dummy value)
        output = [1]  # will be overwritten during training (dummy value)
        hidden = self.options["hidden_layer_sizes"]
        layer_sizes = inputs + hidden + output
        self.model = NeuralNet(layer_sizes)

    def _train(self):
        X, Y, J = _smt_to_genn(self.training_points)
        n_x = X.shape[0]
        n_y = Y.shape[0]
        hidden = self.options["hidden_layer_sizes"]
        layer_sizes = [n_x] + hidden + [n_y]
        self.model.parameters.layer_sizes = layer_sizes
        self.model.parameters.initialize()
        kwargs = dict(
            is_normalize=self.options["is_normalize"],
            alpha=self.options["alpha"],
            lambd=self.options["lambd"],
            gamma=self.options["gamma"],
            beta1=self.options["beta1"],
            beta2=self.options["beta2"],
            epochs=self.options["num_epochs"],
            batch_size=(
                None
                if self.options["mini_batch_size"] < 0
                else self.options["mini_batch_size"]
            ),
            max_iter=self.options["num_iterations"],
            is_backtracking=self.options["is_backtracking"],
            is_verbose=self.options["is_print"],
            random_state=None if self.options["seed"] < 0 else self.options["seed"],
        )
        self.model.fit(X, Y, J, **kwargs)

    def _predict_values(self, x):
        return self.model.predict(x.T).T

    def _predict_derivatives(self, x, kx):
        return self.model.predict_partials(x.T)[:, kx, :].T



"""
Author: Remi Lafage <remi.lafage@onera.fr>

This package is distributed under New BSD license.
"""

import warnings

import numpy as np

from smt.sampling_methods.sampling_method import SamplingMethod
from smt.surrogate_models.krg_based import KrgBased, MixIntKernelType
from smt.surrogate_models.surrogate_model import SurrogateModel
from smt.utils.checks import ensure_2d_array
from smt.utils.design_space import (
    BaseDesignSpace,
    CategoricalVariable,
    ensure_design_space,
)


class MixedIntegerSamplingMethod(SamplingMethod):
    """
    Sampling method decorator that takes an SMT continuous sampling method and
    cast values according x types specification to implement a sampling method
    handling integer (ORD) or categorical (ENUM) features
    """

    def __init__(self, sampling_method_class, design_space, **kwargs):
        """
        Parameters
        ----------
        sampling_method_class: class name
            SMT sampling method class
        design_space: BaseDesignSpace
            design space definition
        kwargs: options of the given sampling method
            options used to instanciate the SMT sampling method
            with the additional 'output_in_folded_space' boolean option
            specifying if doe output should be in folded space (enum indexes)
            or not (enum masks)
        """
        self._design_space = design_space
        if "random_state" in kwargs:
            self._design_space.random_state = kwargs["random_state"]
        elif self._design_space.random_state is None:
            self._design_space.random_state = 42
        self._unfolded_xlimits = design_space.get_unfolded_num_bounds()
        self._output_in_folded_space = kwargs.get("output_in_folded_space", True)
        kwargs.pop("output_in_folded_space", None)
        self._sampling_method = sampling_method_class(
            xlimits=self._unfolded_xlimits, **kwargs
        )
        super().__init__(xlimits=self._unfolded_xlimits)

    def _compute(self, nt, return_is_acting=False):
        x_doe, is_acting = self._design_space.sample_valid_x(
            nt,
            unfolded=not self._output_in_folded_space,
            random_state=self._design_space.random_state,
        )
        if return_is_acting:
            return x_doe, is_acting
        else:
            return x_doe

    def __call__(self, nt, return_is_acting=False):
        return self._compute(nt, return_is_acting)

    def expand_lhs(self, x, nt, method="basic"):
        doe = self._sampling_method(nt)
        x_doe, _ = self._design_space.correct_get_acting(doe)
        if self._output_in_folded_space:
            x_doe, _ = self._design_space.fold_x(x_doe)
        return x_doe


class MixedIntegerSurrogateModel(SurrogateModel):
    """
    Surrogate model (not Kriging) decorator that takes an SMT continuous surrogate model and
    cast values according x types specification to implement a surrogate model
    handling integer (ORD) or categorical (ENUM) features
    """

    def __init__(
        self,
        design_space,
        surrogate,
        input_in_folded_space=True,
    ):
        """
        Parameters
        ----------
        design_space: BaseDesignSpace
            design space definition
        surrogate: SMT surrogate model (not Kriging)
            instance of a SMT surrogate model
        input_in_folded_space: bool
            whether x data are in given in folded space (enum indexes) or not (enum masks)
        categorical_kernel: string
            the kernel to use for categorical inputs. Only for non continuous Kriging.
        """
        super().__init__()
        self._surrogate = surrogate
        if isinstance(self._surrogate, KrgBased):
            raise ValueError(
                "Using MixedIntegerSurrogateModel integer model with "
                + str(self._surrogate.name)
                + " is not supported. Please use MixedIntegerKrigingModel instead."
            )
        self.design_space = ensure_design_space(design_space=design_space)

        self._input_in_folded_space = input_in_folded_space
        self.supports = self._surrogate.supports
        self.options["print_global"] = False
        if "poly" in self._surrogate.options:
            if self._surrogate.options["poly"] != "constant":
                raise ValueError("constant regression must be used with mixed integer")

    @property
    def name(self):
        return "MixedInteger" + self._surrogate.name

    def _initialize(self):
        self.supports["derivatives"] = False

    def set_training_values(self, xt, yt, name=None) -> None:
        xt = ensure_2d_array(xt, "xt")

        # Round inputs
        design_space = self.design_space
        xt, _ = design_space.correct_get_acting(xt)

        if self._input_in_folded_space:
            xt_apply, _ = design_space.unfold_x(xt)
        else:
            xt_apply = xt

        super().set_training_values(xt_apply, yt, name)
        self._surrogate.set_training_values(xt_apply, yt, name)

    def update_training_values(self, yt, name=None):
        super().update_training_values(yt, name)
        self._surrogate.update_training_values(yt, name)

    def _train(self):
        self._surrogate._train()

    def predict_values(self, x: np.ndarray) -> np.ndarray:
        x_corr, is_acting = self._get_x_for_surrogate_model(x)
        return self._surrogate.predict_values(x_corr)

    def _predict_intermediate_values(self, x: np.ndarray, lvl) -> np.ndarray:
        x_corr, is_acting = self._get_x_for_surrogate_model(x)
        return self._surrogate._predict_intermediate_values(x_corr, lvl)

    def predict_variances(self, x: np.ndarray) -> np.ndarray:
        x_corr, is_acting = self._get_x_for_surrogate_model(x)
        return self._surrogate.predict_variances(x_corr)

    def predict_variances_all_levels(self, x: np.ndarray) -> np.ndarray:
        x_corr, is_acting = self._get_x_for_surrogate_model(x)
        return self._surrogate.predict_variances_all_levels(x_corr)

    def _get_x_for_surrogate_model(self, x):
        xp = ensure_2d_array(x, "xp")

        x_corr, is_acting = self.design_space.correct_get_acting(xp)
        if self._input_in_folded_space:
            x_corr, is_acting = self.design_space.unfold_x(x_corr, is_acting=is_acting)
        return x_corr, is_acting

    def _predict_values(self, x: np.ndarray) -> np.ndarray:
        pass


class MixedIntegerKrigingModel(KrgBased):
    """
    Kriging model decorator that takes an SMT continuous surrogate model and
    cast values according x types specification to implement a surrogate model
    handling integer (ORD) or categorical (ENUM) features
    """

    def __init__(
        self,
        surrogate,
        input_in_folded_space=True,
    ):
        """
        Parameters
        ----------
        surrogate: SMT Kriging surrogate model
            instance of a SMT Kriging surrogate model
        """
        super().__init__()
        self._surrogate = surrogate
        if not (isinstance(self._surrogate, KrgBased)):
            raise ValueError(
                "Using MixedIntegerKrigingModel integer model with "
                + str(self._surrogate.name)
                + " is not supported. Please use MixedIntegerSurrogateModel instead."
            )
        self.options["design_space"] = self._surrogate.design_space
        if surrogate.options["hyper_opt"] == "TNC":
            warnings.warn(
                "TNC not available yet for mixed integer handling. Switching to Cobyla"
            )

        self._surrogate.options["hyper_opt"] = "Cobyla"

        self._input_in_folded_space = input_in_folded_space
        self.supports = self._surrogate.supports
        self.options["print_global"] = False

        if "poly" in self._surrogate.options:
            if self._surrogate.options["poly"] != "constant":
                raise ValueError("constant regression must be used with mixed integer")

        design_space = self.design_space
        if (
            any(
                isinstance(dv, CategoricalVariable)
                for dv in design_space.design_variables
            )
            and self._surrogate.options["categorical_kernel"] is None
        ):
            self._surrogate.options["categorical_kernel"] = (
                MixIntKernelType.HOMO_HSPHERE
            )
            warnings.warn(
                "Using MixedIntegerSurrogateModel integer model with Continuous Relaxation is not supported. \
                    Switched to homoscedastic hypersphere kernel instead."
            )
        if self._surrogate.options["categorical_kernel"] is not None:
            self._input_in_folded_space = False

    @property
    def name(self):
        return "MixedInteger" + self._surrogate.name

    def _initialize(self):
        super()._initialize()
        self.supports["derivatives"] = False

    def set_training_values(self, xt, yt, name=None, is_acting=None):
        xt = ensure_2d_array(xt, "xt")

        # If the is_acting matrix is not given, assume input is not corrected (rounding, imputation, etc.) yet
        design_space = self.design_space
        if is_acting is None:
            xt, is_acting = design_space.correct_get_acting(xt)

        if self._input_in_folded_space:
            xt_apply, is_acting_apply = design_space.unfold_x(xt, is_acting)
        else:
            xt_apply, is_acting_apply = xt, is_acting

        super().set_training_values(xt_apply, yt, name, is_acting=is_acting_apply)
        self._surrogate.set_training_values(
            xt_apply, yt, name, is_acting=is_acting_apply
        )

    def update_training_values(self, yt, name=None):
        super().update_training_values(yt, name)
        self._surrogate.update_training_values(yt, name)

    def _train(self):
        self._surrogate._train()

    def predict_values(self, x: np.ndarray, is_acting=None) -> np.ndarray:
        x_corr, is_acting = self._get_x_for_surrogate_model(np.copy(x))
        return self._surrogate.predict_values(x_corr, is_acting=is_acting)

    def _predict_intermediate_values(
        self, x: np.ndarray, lvl, is_acting=None
    ) -> np.ndarray:
        x_corr, is_acting = self._get_x_for_surrogate_model(x)
        return self._surrogate._predict_intermediate_values(
            x_corr, lvl, is_acting=is_acting
        )

    def predict_variances(self, x: np.ndarray, is_acting=None) -> np.ndarray:
        x_corr, is_acting = self._get_x_for_surrogate_model(np.copy(x))
        return self._surrogate.predict_variances(x_corr, is_acting=is_acting)

    def predict_variances_all_levels(self, x: np.ndarray, is_acting=None) -> np.ndarray:
        x_corr, is_acting = self._get_x_for_surrogate_model(x)
        return self._surrogate.predict_variances_all_levels(x_corr, is_acting=is_acting)

    def _get_x_for_surrogate_model(self, x):
        xp = ensure_2d_array(x, "xp")

        x_corr, is_acting = self.design_space.correct_get_acting(xp)
        if self._input_in_folded_space:
            x_corr, is_acting = self.design_space.unfold_x(x_corr, is_acting=is_acting)
        return x_corr, is_acting

    def _predict_values(self, x: np.ndarray, is_acting=None) -> np.ndarray:
        pass


class MixedIntegerContext(object):
    """
    Class which acts as sampling method and surrogate model factory
    to handle integer and categorical variables consistently.
    """

    def __init__(self, design_space, work_in_folded_space=True):
        """
        Parameters
        ----------
        design_space: BaseDesignSpace
            the design space definition (includes mixed-discrete and/or hierarchical specifications)
        work_in_folded_space: bool
            whether x data are in given in folded space (enum indexes) or not (enum masks)
        """

        self._design_space = ensure_design_space(design_space=design_space)
        self._unfold_space = not work_in_folded_space
        self._unfolded_xlimits = self._design_space.get_unfolded_num_bounds()
        self._work_in_folded_space = work_in_folded_space

    @property
    def design_space(self) -> BaseDesignSpace:
        return self._design_space

    def build_sampling_method(self, random_state=None):
        """
        Build Mixed Integer LHS ESE sampler.
        """
        return_folded = self._work_in_folded_space

        def sample(n):
            x, _ = self._design_space.sample_valid_x(
                n, unfolded=not return_folded, random_state=random_state
            )
            return x

        return sample

    def build_kriging_model(self, surrogate):
        """
        Build MixedIntegerKrigingModel from given SMT surrogate model.
        """
        surrogate.options["design_space"] = self._design_space

        if surrogate.options["hyper_opt"] == "TNC":
            warnings.warn(
                "TNC not available yet for mixed integer handling. Switching to Cobyla"
            )

        surrogate.options["hyper_opt"] = "Cobyla"
        return MixedIntegerKrigingModel(
            surrogate=surrogate,
            input_in_folded_space=self._work_in_folded_space,
        )

    def build_surrogate_model(self, surrogate):
        """
        Build MixedIntegerKrigingModel from given SMT surrogate model.
        """
        return MixedIntegerSurrogateModel(
            self._design_space,
            surrogate=surrogate,
            input_in_folded_space=self._work_in_folded_space,
        )

    def get_unfolded_xlimits(self):
        """
        Returns relaxed xlimits
        Each level of an enumerate gives a new continuous dimension in [0, 1].
        Each integer dimensions are relaxed continuously.
        """

        return self._unfolded_xlimits

    def get_unfolded_dimension(self):
        """
        Returns x dimension (int) taking into account  unfolded categorical features
        """

        return len(self._unfolded_xlimits)



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.
"""

import os
import sys


class Silence(object):
    def __enter__(self):
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, *args):
        sys.stdout.close()
        sys.stdout = sys.__stdout__
        return False


class Silence2:
    """Context manager which uses low-level file descriptors to suppress
    output to stdout/stderr, optionally redirecting to the named file(s).

    >>> import sys, numpy.f2py
    >>> # build a test fortran extension module with F2PY
    ...
    >>> with open('hellofortran.f', 'w') as f:
    ...     f.write('''\
    ...       integer function foo (n)
    ...           integer n
    ...           print *, "Hello from Fortran!"
    ...           print *, "n = ", n
    ...           foo = n
    ...       end
    ...       ''')
    ...
    >>> sys.argv = ['f2py', '-c', '-m', 'hellofortran', 'hellofortran.f']
    >>> with Silence():
    ...     # assuming this succeeds, since output is suppressed
    ...     numpy.f2py.main()
    ...
    >>> import hellofortran
    >>> foo = hellofortran.foo(1)
     Hello from Fortran!
     n =  1
    >>> print "Before silence"
    Before silence
    >>> with Silence(stdout='output.txt', mode='w'):
    ...     print "Hello from Python!"
    ...     bar = hellofortran.foo(2)
    ...     with Silence():
    ...         print "This will fall on deaf ears"
    ...         baz = hellofortran.foo(3)
    ...     print "Goodbye from Python!"
    ...
    ...
    >>> print "After silence"
    After silence
    >>> # ... do some other stuff ...
    ...
    >>> with Silence(stderr='output.txt', mode='a'):
    ...     # appending to existing file
    ...     print >> sys.stderr, "Hello from stderr"
    ...     print "Stdout redirected to os.devnull"
    ...
    ...
    >>> # check the redirected output
    ...
    >>> with open('output.txt', 'r') as f:
    ...     print "=== contents of 'output.txt' ==="
    ...     print f.read()
    ...     print "================================"
    ...
    === contents of 'output.txt' ===
    Hello from Python!
     Hello from Fortran!
     n =  2
    Goodbye from Python!
    Hello from stderr

    ================================
    >>> foo, bar, baz
    (1, 2, 3)
    >>>

    """

    def __init__(self, stdout=os.devnull, stderr=os.devnull, mode="wb"):
        self.outfiles = stdout, stderr
        self.combine = stdout == stderr
        self.mode = mode

    def __enter__(self):
        self.sys = sys
        # save previous stdout/stderr
        self.saved_streams = saved_streams = sys.__stdout__, sys.__stderr__
        self.fds = fds = [s.fileno() for s in saved_streams]
        self.saved_fds = map(os.dup, fds)
        # flush any pending output
        for s in saved_streams:
            s.flush()

        # open surrogate files
        if self.combine:
            null_streams = [open(self.outfiles[0], self.mode, 0)] * 2
            if self.outfiles[0] != os.devnull:
                # disable buffering so output is merged immediately
                sys.stdout, sys.stderr = map(os.fdopen, fds, ["w"] * 2, [0] * 2)
        else:
            null_streams = [open(f, self.mode, 0) for f in self.outfiles]
        self.null_fds = null_fds = [s.fileno() for s in null_streams]
        self.null_streams = null_streams

        # overwrite file objects and low-level file descriptors
        map(os.dup2, null_fds, fds)

    def __exit__(self, *args):
        sys = self.sys
        # flush any pending output
        for s in self.saved_streams:
            s.flush()
        # restore original streams and file descriptors
        map(os.dup2, self.saved_fds, self.fds)
        sys.stdout, sys.stderr = self.saved_streams
        # clean up
        for s in self.null_streams:
            s.close()
        return False


## end of http://code.activestate.com/recipes/577564/ }}}



from .ego import EGO, Evaluator
from .mfk import MFK, NestedLHS
from .mfkpls import MFKPLS
from .mfkplsk import MFKPLSK
from .moe import MOE, MOESurrogateModel
from .vfm import VFM
from .podi import PODI

__all__ = [
    "VFM",
    "MOE",
    "MOESurrogateModel",
    "MFK",
    "NestedLHS",
    "MFKPLS",
    "MFKPLSK",
    "EGO",
    "Evaluator",
    "PODI",
]
